#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import json
import re
from collections import defaultdict, Counter
from datetime import datetime
from dateutil.relativedelta import relativedelta

import matplotlib.pyplot as plt
import matplotlib.ticker as tick
import matplotlib.dates as mdates
from matplotlib.sankey import Sankey
import pandas as pd
try:
    import seaborn as sns
except:
    print("Ned to pip install 'seaborn'")
    raise SystemExit("Missing Required Package")
    
"""
TODO NOW:
- simple overview of when bar, boxplot, scatter + shows seaborn (simple) of all
  - https://towardsdatascience.com/a-beginners-guide-to-plotting-your-data-python-r-11e435262ac1 [review to tighten own]
- style: https://medium.com/analytics-vidhya/drastically-beautifying-visualizations-with-one-line-styling-plots-35a5712c4f54 <------ shows built in plt styles (not just seaborn) and can invoke easily. [FRIDAY] ex/ 538 or etc
  - see how these style files created (plt.style.use = https://matplotlib.org/3.3.0/tutorials/introductory/customizing.html) ... take one and then work own over months
- and review overall: http://www.storytellingwithdata.com/

Playfair's bars
- its orientation (vertical and horizontal bar charts) [H or V]
- and its arrangement (e.g. grouped and stacked bar charts) [G or S] 
ie/ denote types ... and variations of all these bars
"""

"""    
> vis = Visualize()
> vis.isHasBH ... etc

Modeled on sns in particular and its one liners. 
    
Using: Matplotlib, Seaborn and Pandas (DF), Numpy for static, PDF/printable graphics

REM: numpy and its expansion of Python's array to multi-dimensional arrays (see DF interplay)

Later, once static works, will try plotDy: web, dynamic versions (see: https://towardsdatascience.com/pyviz-simplifying-the-data-visualisation-process-in-python-1b6d2cb728f1, https://towardsdatascience.com/plotly-dash-vs-streamlit-which-is-the-best-library-for-building-data-dashboard-web-apps-97d7c98b938c )

Refine based on Medium etc articles.

Specific FM data context: [1] complete VISN 20 centric picture within and between VistAs + [2] size and shape of VistA data.

TODO:
- refine/maximize matplotlib
- get more into seaborn (nice wrapper on matplotlib)
- dataframes (see webReportDirectoriesLess)
  - filling/dropping
  - https://towardsdatascience.com/do-you-know-python-has-a-built-in-database-d553989c87bd has to/from sql ... see other formats
- others (more interactive, maps etc): plotly, Altair
  - geopandas and must I go plotly? https://medium.com/towards-artificial-intelligence/data-visualisation-using-pandas-and-plotly-970df88fba6f

Approach to code: this is THE GENERIC VERSION. Reports may have custom plotrs too in a separate module - ex/ plotUsers.py etc.

BACKGROUND (on the "art"):
===========================

> We have Playfair to thank for many of the popular graphs that we use today, including the bar graph (why pies bad: http://www.perceptualedge.com/articles/08-21-07.pdf) + get posters.
   - https://www.amazon.com/Playfairs-Commercial-Political-Statistical-Breviary/dp/0521855543
[note Priestly did first timeline charts: https://en.wikipedia.org/wiki/William_Playfair] 

but 

> Edward Tufte once said that “the only worse design than a pie chart is several of them, for then the viewer is asked to compare quantities located in spatial disarray both within and between pies” ... see ([save pies for the desert](http://www.perceptualedge.com/articles/08-21-07.pdf))
    
This [article](https://medium.com/nightingale/3rd-wave-data-visualization-824c5dc84967) goes into the first (Tufte) and Second (D3 et al) waves with references to other essays on forms and approaches. TODO <------- follow it around and people like:
    - https://twitter.com/jburnmurdoch
    
Visuals to Add:
================
- Radial (for time series) ... by year ... relative additions ... https://medium.com/nightingale/from-the-battlefield-to-basketball-a-data-visualization-journey-with-florence-nightingale-c39571686dfc
  and https://www.pythonprogramming.in/plot-polar-graph-in-matplotlib.html [polar]
  ... can see max ever and then each year within it and then each system in that
- TreeMap (squares size wanted for Workload per site and then relative
size of each square for each clinic/entity?)
  - import plotly.graph_objects as go / fig = go.Figure(go.Treemap
  
  
Not much magic here. We import lru_cache and decorate a function that generates Fibonacci numbers with it.

    @lru_cache(maxsize=None)
    def do(x):  

"""
class Visualize:

    def __init__(self, flushToDirectory=""): 
        self.__flushToDirectory = flushToDirectory or "Images"
        
    def svgPlotFile(self, plotName, plotMethodName): # rem: kargs inside plotName
        """
        So don't have to make the plot but can know its name as making md
        Compatible with __flushPlot's naming of SVG's
        """
        coreMethodName = re.sub("plot", "", plotMethodName)
        coreMethodNameU = coreMethodName[0].upper() + coreMethodName[1:]
        plotFilePrefix = f'{plotName}{coreMethodNameU}'
        plotFile = f'{self.__flushToDirectory}/{plotFilePrefix}BGW.svg'
        return plotFile
        
    def makeDF(self, data, columns, rows):
        """
        Note: 'rows' == 'index' == first column values stuffed in columns
        
        REM: index could be a column (again) if you .reset_index(inplace=True)
        
        Ex/ 
        
            data = [ (aka rows)
                ("val col1", "val col2", "val col3") # of row 1
                ("val col1", "val col2", "val col3") # of row 2
            ]
            columns: ["col1 name", "col2 name", "col3 name"],
            rows: ["row1 name", "row2 name"] (aka rowLabels)
        
        """
        df = pd.DataFrame(
            data, 
            columns=columns,
            index=rows
        )
        return df
        
    def plotIsHasBH(self, countsNTotal, title, plotName, usePerc=True):
        plotIsHasBH(countsNTotal, usePerc) # title is ignored
        plotFilePrefix = f'{plotName}IsHasBH' # add gr type to prefix
        return self.__flushPlots(plotFilePrefix)
                
    def plotCategoryBH(self, df, title, plotName):
        plotCategoryBH(df, title)
        plotFilePrefix = f'{plotName}CategoryBH' # add gr type to prefix
        return self.__flushPlots(plotFilePrefix)
        
    def plotOutOtherSankey(self, specs, title, plotName):
        plotOutOtherSankey(specs["outValues"], specs["toOtherTTL"], specs["fromLabel"], title)
        plotFilePrefix = f'{plotName}OutOtherSankey' # add gr type to prefix        
        return self.__flushPlots(plotFilePrefix)
        
    # Two variations - absolute and usePerc
    def plotCategoryBSV(self, totalsByCatagByStackId, title, plotName, usePerc=False):
        plotCategoryBSV(totalsByCatagByStackId, title, usePerc)
        plotFilePrefix = f'{plotName}CategoryBSV'    
        return self.__flushPlots(plotFilePrefix)
                        
    def plotDailies(self, dailies, title, plotName, colNameById, start='2020-01', end='2020-02'):
        plotDailies(dailies, title, colNameById, start, end)
        plotFilePrefix = f'{plotName}Dailies{start}-{end}' # add gr type to prefix
        return self.__flushPlots(plotFilePrefix)

    def plot365RollingMean(self, dailies, title, plotName, colNameById, start="1987-01", end=""):
        plot365RollingMean(dailies, title, colNameById, start, end)
        plotFilePrefix = f'{plotName}365RollingMean{start}-{end}' # add gr type to prefix
        return self.__flushPlots(plotFilePrefix)

    def plotMonthlyResamples(self, dailies, title, plotName, colNameById, start='2018-01', end='2020-02'):
        plotMonthlyResamples(dailies, title, colNameById, start, end)
        plotFilePrefix = f'{plotName}MonthlyResamples{start}-{end}' # add gr type to prefix
        return self.__flushPlots(plotFilePrefix)

    def plotTotalsPerYear(self, dailies, title, plotName, colNameById, start="2015", end="2020", snosOnly=True):
        plotTotalsPerYear(dailies, title, colNameById, start, end, snosOnly)
        plotFilePrefix = f'{plotName}TotalsPerYear{start}-{end}' # add gr type to prefix
        return self.__flushPlots(plotFilePrefix)

    def __flushPlots(self, plotFilePrefix):
        """
        Four variations - may move to two transparent (svg and png)

        # TODO: see the grid color settings - default is white which would disappear in
        # a transparent to white bg. So need to work grid darker but fine?
        # ... 
        #
        # ? bbox_inches="tight" as extra or rely on explicit above?
        """
        fullPlotFilePrefix = f'{self.__flushToDirectory}/{plotFilePrefix}'
        
        plt.savefig(f'{fullPlotFilePrefix}.png', transparent=True)
        plt.savefig(f'{fullPlotFilePrefix}BGW.png', transparent=False)
        fullBGWSVGFile = f'{fullPlotFilePrefix}BGW.svg'
        try:
            plt.savefig(f'{fullPlotFilePrefix}.svg', transparent=True, format="svg")
            plt.savefig(f'{fullBGWSVGFile}', transparent=False, format="svg")
        except:
            print(f'Flushed 2 images \'{fullPlotFilePrefix}...png\' - svg failed (Redhat?)')    
        else:
            print(f'Flushed 4 images \'{fullPlotFilePrefix}...png/svg\'')
        return fullBGWSVGFile
        
# ----------------------- Individual Plots -------------------------#

"""
Individual plots made but NOT saved or shown - plt

To use:
    plotX(args)
Then either:
    plt.show()
        or
    plt.savefig("to...", ...)
"""
        
# ####################### Stacked Bar x 3 ###########################

def plotIsHasBH(countsNTotal, usePerc=True):
    """
    IsHas (one abs value out of total) - Bar Horizontal
    
    For:
    - [1] comparison of attributes (ex/ is Male, is Deceased, ...) for a total (ex/ 
    patients) of one entity (ex/ patient directory). Attributes are the Y Axis
    and the LVALUE of the dictionary passed in.
    - [2] side by side comparison of percent of one attribute (ex/ Male)
    for many entities (ex/ VistA 1, VistA 2 ...). Entities are the Y Axis and the 
    LVALUE of the dictionary passed in.
    but not for category/enum information (ex/ Male vs Female, ALL vs Some VistAs)
    
    Options: usePerc - use percentages vs usePerc = False => absolutes shown
        
        {
            "__entityName": ...
            "__total": #,
            "assertion 1": # <= total, (ex/ in one vista, sent to other sites ...)
            ...
        }
        
    The Percents are calculated inside this function. The #'s passed are absolutes. The
    attributes are reordered by size.
    
    Note: alternative often used is a DL with form
    
        blah blah ..
                  <---- newline
            assertion 1
             : %  <---- one space, colon
                  <---- newline til next entry
            assertion 2
             : %
    
    TODO (cosmetic)
    - see if can do a DF perc (see tip at bottom of this: 
    https://python-graph-gallery.com/13-percent-stacked-barplot/ with axis etc)
    - ala WSJ
      - X bar on top (PERC or ABS) ... perc -- only 0% has it/ 5 / 10 etc.
      - left align labels
      - white BG
      - no horiz lines (got this); verticals light grey not white so show
    """
    if not ("__entityName" in countsNTotal and "__total" in countsNTotal):
        raise Exception("'countsNTotal' needs __entityName and __total")
    name = countsNTotal["__entityName"]
    total = countsNTotal["__total"]
    index = []
    data = {"SET": [], "NOTSET": []}
    for ya in sorted([x for x in countsNTotal if not re.match(r'__', x)], key=lambda x: countsNTotal[x]): 
        if not isinstance(countsNTotal[ya], int):
            raise Exception("must be a simple count")
        index.append(ya)
        if usePerc:
            setValue = 100 * countsNTotal[ya]/total
            notSetValue = 100 - setValue
        else:
            setValue = countsNTotal[ya]
            notSetValue = total - countsNTotal[ya]
        data["SET"].append(setValue)
        data["NOTSET"].append(notSetValue)

    sns.set() 
    fig, ax = plt.subplots(figsize=(11, 4)) # fixed 11/4
    ax.set_title("{:,} {}".format(total, name.capitalize()))
    width = .75 # the width of the bars: can also be len(x) sequence
    ax.barh(index, data["SET"], width, edgecolor="none")
    ax.barh(index, data["NOTSET"], width, left=data["SET"], color="lightgray", edgecolor="none")
    ax.xaxis.set_major_formatter(plt.NullFormatter())
    # ax.xaxis.set_major_formatter(tick.FuncFormatter(reformat_large_tick_values));
    ax.margins(0) # this gets rid of the right spine's extra margin (ax.spines["right"])
    ax.yaxis.set_tick_params(pad=15) # note: couldn't bold em easily ... LaTeX TODO
    ax.tick_params(axis='both', which='both',length=0)
    for i, x in enumerate(data["SET"]):
        if usePerc:
            if x < 90:
                xpos = x + 1.25
            else:
                xpos = x - 5
            valToShow = "{:,}%".format(round(x, 1))
        else:
            if x/total < 0.90:
                xpos = x + total/100
            else:
                xpos = x - total/12
            valToShow = "{:,}".format(reformat_large_tick_values(x))
        plt.text(xpos, i, valToShow, fontsize=11, fontweight="bold", va='center')  
    plt.tight_layout()
    
def plotCategoryBH(df, title):
    """
    Category (vs isHas) Bar Horizontal for different items with the same 
    category breakdowns. Alternative to pie charts.
    
                       title (tip: put a # message in the title)
    
    rowLabel1    |    %col2         | %col2     | %col3 | #
    
    rowLabel2    |
   
    where key will be col1, col2, col3 and the rows have data with the same breakdown (col1, ...) but can be different orders ie/ total up differently. Ex/ records and files.
           
    Accepts df. Ex/
        rowLabels = ["rowLabel1", "rowLabel2", "rowLabel3", "rowLabel4"]
        cols = ["col1", "col2", "col3"]
        rows = [(89361, 5907, 0), (262, 209, 0), (390, 336, 44), (13151, 3454, 500)]    
        df = pd.DataFrame(
            rows,
            columns=cols,
            index=rowLabels
        )
    Will turn into percentages.
    (Tip: df = df.drop("services"))

    TODO more: 
    - see ... more on score inside 
    - consider % on right if #'s don't all total to the same ie/ ala my % of % tables
    as this is a replacement and also order down on those? ie/ lessor total % in lower
    rows
    
    REF:    https://matplotlib.org/3.1.1/gallery/lines_bars_and_markers/horizontal_barchart_distribution.html#sphx-glr-gallery-lines-bars-and-markers-horizontal-barchart-distribution-py <--- NOT this method but the look side to side is ok
    
    """        

    # perc version used for all but end total!
    dfp = df.div(df.sum(axis=1), axis=0).mul(100).round(3) 
        
    sns.set() 
    figsize = (9, 2 * len(df.index))
    fig, ax = plt.subplots(figsize=figsize)
    ax.margins(0) # this gets rid of the right spine's extra margin (ax.spines["right"])
    ax.yaxis.set_tick_params(pad=15) # note: couldn't bold em easily ... LaTeX TODO
    ax.xaxis.set_visible(False)
    if len(dfp[df.columns[0]].values) == 1:
        ax.yaxis.set_visible(False)
    ax.set_title(title)
    
    width = .75 # the width of the bars: can also be len(x) sequence
    """
    Col by Col added across all y's (df.index) with shift right each time
        y: df.index (all the "first col" values), 
        x: dfp[colName].values (this cols values)
        label == y label ie/ of horizontal bar
    REM: y first as barh
    
    Alt would be put index (first col) into rows at pos 0 and always give its
    values and then values of next col's index in all rows but better to keep
    columns (index) separate
    
    or TODO: iterrows (as just in => cols only)
    
        con = []
        for i, row in df.iterrows():
            g = {‘Customer ID #: ‘+str(row[‘customer_id’]): 
              {‘Address’: row[‘address’],
               ‘Post Code’: row[‘postcode’],
               ‘State’: row[‘state’], 
               ‘Country’: row[‘country’]
            }}
            con.append(g)
    """    
    for i, colName in enumerate(dfp.columns):
        # TODO: want first on top -- must fix
        # ys = list(reversed(df.index)) # as horiz - want first row on top
        if i == 0:
            ax.barh(df.index, dfp[colName].values, width, label=colName)
            ttlsSoFar = dfp[colName].values[:]
            continue
        ax.barh(df.index, dfp[colName].values, width, left=ttlsSoFar, label=colName)
        ttlsSoFar = [a + b for a, b in zip(ttlsSoFar, dfp[colName].values)]
    ax.legend(fontsize=10, loc='upper left')
    
    rowValSoFarByRowIdx = Counter()
    for colName in df.columns: # go across
        for i, val in enumerate(dfp[colName]): # go down
            rowValSoFarByRowIdx[i] += val
            if val < 3:
                continue
            display = f'{round(val)}%'
            xoffset = round(len(display)/1.25)
            xForVAL = rowValSoFarByRowIdx[i] - (round(val/2) + xoffset)
            plt.text(xForVAL, i, display, fontsize=11, fontweight='bold', va='center')
    ttls = df.sum(axis=1)
    for i, ttl in enumerate(ttls):
        ttlToShow = reformat_large_tick_values(ttl)
        x = 100 + round(figsize[0] * 0.25)
        plt.text(x, i, ttlToShow, fontsize=11, fontweight='bold', va='center')
    plt.tight_layout()
    
"""
Classic Vertical Stackbar (BSV) for a category breakdown, total or same size percentages. 

TODO MORE WHEN TIME: 
> plt.rcParams.update({'font.size': 22})
and
params = {'legend.fontsize': 'large',
          'figure.figsize': (20,8),
          'axes.labelsize': size,
          'axes.titlesize': size,
          'xtick.labelsize': size*0.75,
          'ytick.labelsize': size*0.75,
          'axes.titlepad': 25}
plt.rcParams.update(params)
ie/ calculate and set all

TMP FIRST: just totals ...

        totalsByCatagByStackId = {
            "WWW": {
                NONE: 7499,
                2: 7802,
                3: 27625,
                ALL: 177852
            },
            "SPO": {
                NONE: 20478,
                2: 33649,
                3: 92134,
                ALL: 177852
            },
            ...
        }

TODO:
- GENERALIZE (ex/ https://medium.com/swlh/converting-nested-json-structures-to-pandas-dataframes-e8106c59976e + https://python-graph-gallery.com/13-percent-stacked-barplot/ comments)
- df
- break in two ... 
- option to side by side (extra wrapper)

https://matplotlib.org/gallery/lines_bars_and_markers/bar_stacked.html#sphx-glr-gallery-lines-bars-and-markers-bar-stacked-py

Key TODO for this and the totals above is to move to DataFrame proper 

    # TODO: how to make DF do the percent 
    df = pd.DataFrame(data, index)
    df[1].value_counts()
    print(df.head())
    print(data)
    print(index)
    
ie/ want DF from data, index and then [1] percentage from base one (don't 
manually calc) [2] as below for TimeSeries, drive off DF
"""
def plotCategoryBSV(totalsByCatagByStackId, title, usePerc=False): # bar, stacked, vertical

    """
    matplotlib direct: https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.bar.html#matplotlib.axes.Axes.bar
    pandas.dataframes.plot (kind="barh")
    """
        
    """
    Turn StackId into index and catag "owns" dictionary ie/
        www: {"ALL": # ---> {"ALL: [# for WWW, ...
    which lets us paint the graph layer by layer from bottom to top
    
    X axis (index) of VistA (SNO) (vs date in time series)
    Y axis as int amounts (same as time series)
    Legend/color/column shown to be 1, 2, 3, 4 (ala VistA (SNO) in time series)
    """
    # want in order of size, smallest to the right
    index = []
    data = defaultdict(list) # index is legend/Y ie/ ALL, 3, ...; X is list index
    masterCatagOrder = [] # ensuring stacks have same order
    for i, stackId in enumerate(sorted(totalsByCatagByStackId, key=lambda x: sum(totalsByCatagByStackId[x][k] for k in totalsByCatagByStackId[x]))):
        index.append(stackId)
        stackTTL = sum(totalsByCatagByStackId[stackId][catag] for catag in totalsByCatagByStackId[stackId])
        for j, catag in enumerate(totalsByCatagByStackId[stackId]): # assuming ordered
            if i == 0:
                masterCatagOrder.append(catag)
            elif masterCatagOrder[j] != catag:
                raise Exception("Stacks passed in don't have categories in same order")
            # str(catag) as JSON dump turns lvalue ints to strs
            if usePerc == False:
                data[catag].append(totalsByCatagByStackId[stackId][str(catag)])
            else:
                data[catag].append(100 * totalsByCatagByStackId[stackId][str(catag)]/stackTTL)
                                 
    sns.set() 
    fig, ax = plt.subplots(figsize=(len(index) * 1.5, 5)) # NEED TO CALC FROM HEIGHT DIFFERENCE TODO
    if usePerc == False:
        ax.yaxis.set_major_formatter(tick.FuncFormatter(reformat_large_tick_values))

    width = .75 # ie/ leave .125 each side between bars
        
    bars = []
    xs = [-0.15, 0.85, 1.85, 2.85, 3.85, 4.85, 5.85]
    ttlsSoFar = [0 for i in range(0, len(totalsByCatagByStackId))]
    firstTTLs = None
    def formatNumber(fnum):
        num = round(fnum, 1)
        if num % 1 == 0:
            return int(num)
        else:
            return num
    for i, catag in enumerate(data):
        barn = ax.bar(index, data[catag], width, bottom=ttlsSoFar) # layer
        bars.append(barn)
        if usePerc: # % inside bar pieces if > 10%
            for j, dp in enumerate(data[catag]):   
                if dp < 10:
                    continue
                plt.text(xs[j], (ttlsSoFar[j] - 1) + dp/2, f'{formatNumber(dp)}%', fontsize=11) 
        ttlsSoFar = [a + b for a, b in zip(ttlsSoFar, data[catag])] 
        if i == 0:
            firstTTLs = ttlsSoFar[:]
    if not usePerc: # total on top of bar
        for i, ttl in enumerate(ttlsSoFar):
            dataStr = reformat_large_tick_values(ttlsSoFar[i], roundTo=2)
            plt.text(xs[i], 1.01 * ttlsSoFar[i], dataStr, fontsize=10) 
        # Only want text first if > 10%
        if len(set(firstTTLs)) == 1 and firstTTLs[0]/ttlsSoFar[-1] > .1: 
            noToShow = firstTTLs[0]
            txt = plt.text(xs[len(totalsByCatagByStackId) - 1], round(firstTTLs[0] * 0.9), reformat_large_tick_values(firstTTLs[0], roundTo=2), fontsize=11)
            
    legendLoc = "upper left" if usePerc == False else "lower left" # as big % first on left while biggest on right for absolutes
    ax.legend(reversed([barn[0] for barn in bars]), reversed(list(data.keys())), fontsize=9, loc=legendLoc) # ncol=len(bars) setting ncol will make legend horizontal
    # Two ways to relabel - use label= as above; the handles code here; legend map
    # handles, labels = ax.get_legend_handles_labels()
    # ax.legend(reversed(handles), reversed(labels), loc='upper right')
    ax.yaxis.get_major_ticks()[0].label1.set_visible(False) # remove '0'   
    
    # ax.set_ylabel("{} Count".format(dirName.capitalize()))
    ax.set_title(title) # whatever passes as auto needs too many overloads
            
    plt.tight_layout()
    
# ####################### Sankey ####################

"""
TODO:
- % on outs (not absolutes)
- put total into title if want ex/ Travel (X%)
- size --- see if height or width changes make distance better (could use fuller names than BOI etc)
- clean comment below
- see if anything in https://flothesof.github.io/sankey-tutorial-matplotlib.html on angles etc and also on "other"
------
- add a sankey from https://towardsdatascience.com/sankey-diagram-basics-with-pythons-plotly-7a13d557401a (simple and flows better for the many of Travel!)
                and
https://towardsdatascience.com/visualizing-in-app-user-journey-using-sankey-diagrams-in-python-8373a7bb2d22 <------- could do TELER events ... completed transitions! ... show all event flows [for test sets too]

https://het.as.utexas.edu/HET/Software/Matplotlib/api/sankey_api.html

This better tut: https://flothesof.github.io/sankey-tutorial-matplotlib.html

or this https://stackoverflow.com/questions/26677690/connecting-flows-in-matplotlib-sankey-diagram

DO THIS: https://towardsdatascience.com/the-what-why-and-how-of-sankey-diagrams-430cbd4980b5 <---- first

Syrian ref -- total source

(scale matters and as does trunk length ... PERFECT!

AFTER GET THIS CLASS WORKING, do https://github.com/ricklupton/floweaver

and flow from d3 https://www.d3-graph-gallery.com/arc.html

GOAL: get label size right

Mainly based on https://towardsdatascience.com/the-what-why-and-how-of-sankey-diagrams-430cbd4980b5

TODO: 
- WANT AUTO PLACEMENT, scale etc
- remove border

FUTURE:
- ADVANCE - want to combine two ie/ IFC out of VISN

NEXT GO TO IMPROVE SANKEY:
- GITHUB made good normalized versions on top of matplotlib: https://github.com/anazalea/pySankey/blob/master/pysankey/sankey.py

    outValues = [(1355, "PUG"), (680, "POR"), (615, "BOI"), (263, "SPO"), (37, "WCTY")]
    plotOutOtherSankey(outValues, 22, "Other IFCs Placed")
"""
def plotOutOtherSankey(outValues, toOtherTTL, fromLabel, title): 
    
    # Use seaborn style defaults, bg etc
    sns.set() 
    
    outValues = sorted(outValues, key=lambda x: x[0], reverse=True) # big to small north
    outTTL = sum(d[0] for d in outValues)
    
    if outTTL == 0 and toOtherTTL == 0:
        raise Exception(f'Can only OutOther Sankey if there is at least one or other total for OUT or OTHER: {title}')
    
    # outDir = 1 if outTTL > toOtherTTL else -1
    outDir = 1 # keeping one orientation (not differing with size)
    # toOtherDir = 1 if outDir == -1 else -1 
    toOtherDir = -1
    
    flows = [d[0] * -1 for d in outValues] 
    labels = [d[1] for d in outValues]
    orientations = [outDir for d in outValues]
    pathlengths = [0.25 for d in outValues]

    # Only show OTH is there is one!
    if toOtherTTL:
        flows.append(toOtherTTL * -1)
        labels.append("Other VISNs")       
        orientations.append(toOtherDir)
        pathlengths.append(0.25)

    total = outTTL + toOtherTTL
    # labels.insert(0, fromLabel) - don't bother with From as known
    labels.insert(0, "")
    flows.insert(0, total)
    orientations.insert(0, 0)
    pathlengths.insert(0, 0.1)
    
    # Product of scale and total input should be 1 for them but too fat for me
    SCALE_PRODUCT = 0.5  
    scale = SCALE_PRODUCT/total
    
    plt.rcParams['font.size']  = 9 # from https://stackoverflow.com/questions/58900854/sankey-with-matplotlib-positions-of-labels
    
    Sankey(
        scale=scale, # interplay with trunk len as thinkens line
        offset=0.15, # offset of text from arrows - TODO: WWW on left too close still
        # margin=0.4, (around diag in its box, not effect on contents)
        margin=0.3,
        format='%d',
        trunklength = 1, # default is 1
        pathlengths = pathlengths, # default is 0.25
        edgecolor = '#027368', # or white only on smallest!
        facecolor = '#027368',
        flows=flows,
        # labels=["WWW", "PUG", "PORT", "BOISE", "SPOK", "WHITE", "OTHER"],
        labels=labels,
        # pathlengths=[1, 1, 1, 3, 3, 3, 3],
        # patchlabel="Placer Consults",
        orientations=orientations
    ).finish()
    plt.title(title)
    
# ######################## Time Series ########################

"""
Based on https://www.dataquest.io/blog/tutorial-time-series-analysis-with-pandas/
- pandas time series
- matplotlib
- sns (need more info)

Consider weekday out -- 

    from pandas.tseries.offsets import BDay  
    isBusinessDay = BDay().onOffset
    csv_path = 'C:\\Python27\\Lib\\site-packages\\bokeh\\sampledata\\daylight_warsaw_2013.csv'
    dates_df = pd.read_csv(csv_path)
    match_series = pd.to_datetime(dates_df['Date']).map(isBusinessDay)
    dates_df[match_series]
    Crude df[df.index.dayofweek < 5]

TODO: Typer and More
- Fill in gaps in Patient date created data
- Feed to subtyping (needs to be more compact if its to keep dailies!)
  - DatetimeIndex, the data type datetime64[ns] indicates that the underlying data is stored as 64-bit integers, in units of nanoseconds (ns). This data structure allows pandas to compactly store large sequences of date/time values and efficiently perform vectorized operations using NumPy datetime64 arrays.
  ... consider timestamps compacted. What effect on size ... try one as an experiment 
  - down sample ala combining subtypes ... ease of indexing/totaling by month => any datetime byValueCount -- can do one matrix of MONTH: value subtype one, value subtype two etc
    - overall keep creates to day => show below for any type

Not used:
=========

sns' boxplots as no great monthly variation to showcase ...

    fig, axes = plt.subplots(3, 1, figsize=(11, 4), sharex=True)
    for name, ax in zip(['663', '668', '687'], axes):
        sns.boxplot(data=dailies, x='month', y=name, ax=ax)
    ax.set_ylabel('Creations')
    ax.set_title(name)
    # Remove the automatic x-axis label from all but the bottom subplot
    if ax != axes[-1]:
        ax.set_xlabel('')
    plt.show()
    return
    
# TODO: allow end="" ie/ start:"" works or?
"""
def plotDailies(dailies, title, colNameById, start='2020-01', end='2020-02'):

    # Use seaborn style defaults, bg etc
    sns.set() 
       
    fig, ax = plt.subplots(figsize=(11, 6))
    # Hone in further and add a marker and line style
    # ... must plot directly with matplotlib and not the facade as date
    # calculation for the date formatters differs a little and sets markers off
    for colId in colNameById: # not 'columns' as > colId's in cols now
        """
        FIX ABOVE for user dir etc -- TODO not here
        
        if not re.match(r'\d+$', colId):
            continue # exclude ALL but VistAs ... no 'overall' or total users
        """
        dailiesLocCol = dailies.loc[start:end, colId]
        legendLabel = f'{colNameById[colId]} {int(dailiesLocCol.min())}/{int(dailiesLocCol.median())}/{int(dailiesLocCol.max())}'
        ax.plot(dailiesLocCol, marker='o', linestyle='-', label=legendLabel)
    ax.set_title(title)
    ax.set_ylabel('Daily Addition')
    ax.xaxis.set_major_locator(mdates.WeekdayLocator(byweekday=mdates.MO))
    # Format x-tick labels as 3-letter month name and day number (strftime form)
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d')) # '%b %d'
    # ax.xaxis.grid(color='blue', alpha=0.5) -- work the axis color for 
    ax.legend()
        
    plt.tight_layout()
    
"""
365-day rolling mean - note 7-day rolling mean largely follows actual dailies

> An easy way to plot these trends is with rolling means at different time scales
...
> Rolling window operations are another important transformation for time series data. Similar to downsampling, rolling windows split the data into time windows and and the data in each window is aggregated with a function such as mean(), median(), sum() ... unlike downsampling, where the time bins do not overlap and the output is at a lower frequency than the input, rolling windows overlap and “roll” along at the same frequency as the data, so the transformed time series is at the same frequency as the original time series.
"""
def plot365RollingMean(dailies, title, colNameById, start="1987-01", end="2021-12"):
        
    # 365-day rolling mean 
    _365DayRollingMeans = dailies.loc[start:end, list(colNameById)].rolling(window=365, center=True).mean()    
    sns.set() # makes bg and lines
    fig, ax = plt.subplots(figsize=(11, 6))
    for colId in colNameById:
        """
        TODO: fix above for users etc
        
        if not re.match(r'\d+$', sno):
            continue # exclude ALL but VistAs ... no 'overall' or total users
        """
        ax.plot(_365DayRollingMeans[colId], label=colNameById[colId])
    # Set x-ticks to yearly interval, adjust y-axis limits, add legend and labels
    ax.xaxis.set_major_locator(mdates.YearLocator(2))
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%y'))
    ax.legend()
    ax.set_ylabel('Added')
    ax.set_title(title)
    
    plt.tight_layout()

"""
resample the data to monthly frequency, aggregating with sum totals instead of the mean.

Note: article used stacked for some subdata
    opsd_monthly[['Wind', 'Solar']].plot.area(ax=ax, linewidth=0)
"""
def plotMonthlyResamples(dailies, title, colNameById, start='2018-01', end='2020-02'):
    
    # TODO: add check to make sure end gives full month to play with from
    # passed in data (ONLY CHECKING START NOW!)
    
    # TODO: may change to "end" and "timeBack" ala utility => easier to QA inside
    # graph and no duplication logic <----- see timeBack in CSV prod util
    # Must make sure that start is supported by data frame input and prep to resample
    # a month back from asked for support date to avoid dip while also avoiding need
    # to resample the whole of a large data collection
    startDay = start
    if len(startDay) == 4:
        startDay = f'{startDay}-01-01'
    elif len(startDay) == 7:
        startDay = f'{startDay}-01'  
    startDayDT = datetime.strptime(startDay, "%Y-%m-%d")
    startDayLess1MTHDT = startDayDT - relativedelta(months=1)
    startDayLess1MTH = datetime.strftime(startDayLess1MTHDT, "%Y-%m-%d")
    firstRowDay = datetime.strftime(dailies.index[0].to_pydatetime(), "%Y-%m-%d")
    if firstRowDay > startDayLess1MTH:
        raise Exception(f'Can\'t monthly resample from start date less a month, {startDayLess1MTH}, as first data date, {firstRowDay}, comes after that')
    # Resample going back a month to make sure won't start on downturn
    dailies = dailies.loc[startDayLess1MTH:end]

    monthlies = dailies[list(colNameById)].resample('M').sum()
        
    sns.set() # makes bg and lines
    fig, ax = plt.subplots()
    for colId in colNameById: 
        # color='black', green orange red
        ax.plot(monthlies.loc[start:end, colId], label=colNameById[colId])
    # ax.xaxis.set_major_locator(mdates.YearLocator())
    # ax.xaxis.set_major_locator(mdates.WeekdayLocator(byweekday=mdates.MO))
    # Format x-tick labels as 3-letter month name and day number (strftime form)
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%y'))
    ax.legend()
    ax.set_ylabel('Monthly Total')
    ax.set_title(title)
    
    plt.tight_layout()
    
"""
Bar Totals per Year, side by side, by Pandas' wrapper of plt, overridden in places

Possible TODO: merge in automatic vs explicit creation breakdown ie/ BAR with
bottom as used above.

Ex annuals.head() ie/ the resample totals the days to years

       663   648  668  687  USER
Date                            
1984   151     0    0    0   151
1985   481     0    0    0   481
1986  1167     0    0    0  1167
1987  1429  2318  369  113  4222
1988  1342   666  168  118  2281

Plot: X axis is the index (DATE), Y has col values and column name (VistA (SNO)) is legend/color/bar

TODO: add in plt.annotate('Notice something?', xy=('2014-01-01', 30), xytext=('2006-01-01', 50), arrowprops={'facecolor':'red', 'shrink':0.05})
... consider doing manuals like this on top of returned plots => save is a separate
step?
"""
def plotTotalsPerYear(dailies, title, colNameById, start="2015", end="2020", snosOnly=True):

    # TODO: FIX THIS SO NO SNOS etc -- just column names

    # Make sure only get VistA (sno) columns, not the total user | patient
    snos = dict((vnp, colNameById[vnp]) for vnp in colNameById if re.match(r'\d+$', vnp)) if snosOnly else colNameById
    
    dailiesRange = dailies.loc[start:end]
    annuals = dailiesRange[list(snos)].resample('A').sum()
    annuals = annuals.set_index(annuals.index.year)
    annuals.index.name = 'Year'

    sns.set() # makes bg and lines    

    ax = annuals.plot.bar()
    
    ax.set_title(title)
    
    # The following resets pandas plot wrapper defaults (rotated xticks, SNO not name ...)
    ax.legend(snos.values()) # want PUG, SPO etc
    # plt.xticks([0, 1, 2, 3, 4, 5], ["2015", "16", "17", "18", "19", "2020"], rotation=0)
    xtickLabels = [str(yr) if str(yr) in [start, end] else str(yr)[2:] for yr in range(int(start.split("-")[0]), int(end.split("-")[0])+1)]
    ax.set_xticklabels(xtickLabels, rotation=0) # method 2
    ax.set_xlabel("") # don't need to see "Year"
    ax.yaxis.get_major_ticks()[0].label1.set_visible(False) # remove '0'
    
    plt.tight_layout()
    
# ############################## UTILS ########################

def reformat_large_tick_values(tick_val, pos=0, roundTo=1):

    # code below will keep 4.5M as is but change values such as 4.0M to 4M since that zero after the decimal isn't needed
    def removeTrail0(no):
        # make no into a string value
        nos = str(no)
        index_of_decimal = nos.find(".")
        if index_of_decimal == -1:
            return no    
        value_after_decimal = nos[index_of_decimal+1]
        if value_after_decimal == "0":
            # remove the 0 after the decimal point since it's not needed
            nos = nos[0:index_of_decimal]
            return int(nos)
        return no

    """
    Turns large tick values (in the billions, millions and thousands) such as 4500 into 4.5K and also appropriately turns 4000 into 4K (no zero after the decimal).
    """
    if tick_val >= 1000000000:
        val = round(tick_val/1000000000, roundTo)
        new_tick_format = f'{removeTrail0(val)}B'
    elif tick_val >= 1000000:
        val = round(tick_val/1000000, roundTo)
        new_tick_format = f'{removeTrail0(val)}M'
    elif tick_val >= 1000:
        # val = round(tick_val/1000, roundTo)
        # new_tick_format = f'remoteTrail0(val)K'
        new_tick_format = f'{removeTrail0(round(tick_val, 1)):,}'
    elif tick_val < 1000:
        new_tick_format = f'{removeTrail0(round(tick_val, 1))}'
    else:
        new_tick_format = tick_val
            
    return new_tick_format
    
# ############################## TEST DRIVER ################################
    
# TODO: retire/ deprecate so charts off
VISTAS_FNAMES = {"668": "Spokane", "663": "Puget Sound", "648": "Portland", "687": "Walla Walla"}
VISTA_NAMES = {"663": "PUG", "648": "POR", "668": "SPO", "687": "WWW"}
                
def main():

    assert sys.version_info >= (3, 6)

    viz = Visualize("ImagesHere")
    
    rows = ["consults", "services", "users", "patients"]
    columns = ["non ifc", "ifc", "both"]
    data = [(89361, 5907, 0), (262, 209, 0), (390, 336, 44), (13151, 3454, 500)]    
    df = pd.DataFrame( # direct or could do viz.makeDF
        data,
        columns=columns,
        index=rows
    )    
    print(f'Plot would be stored in {viz.svgPlotFile("ifcAndOther", "plotCategoryBH")}')
    plotCategoryBH(df, "IFCs vs Non IFCs", "visualEx")
    plt.show()
    
    outValues = [(1355, "PUG"), (680, "POR"), (615, "BOI"), (263, "SPO"), (37, "WCT")]
    print(f'Plot would be stored in {viz.svgPlotFile("placerOther", "plotOutOtherSankey")}')
    plotOutOtherSankey(outValues, 22, "WWW", "Other IFCs Placed")
    plt.show()
        
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
import statistics

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator, SCHEMA_LOCN_TEMPL, DATA_LOCN_TEMPL, DATARF_LOCN_TEMPL, TMPWORKING_LOCN_TEMPL
from fmqlutils.typer.reduceTypeUtils import checkDataPresent
from fmqlutils.typer.reduceReportType import TYPER_REDUCTIONS_LOCN_TEMPL
from fmqlutils.reporter.reportUtils import MarkdownTable

"""
Site
"""
SITE_DIR_TEMPL = VISTA_DATA_BASE_DIR + "{}/ReportSite/" 

TOP_MD_TEMPL = """---
layout: default
title: {}
---

"""

def ensureWebReportLocations(stationNo):
    reportSiteLocation = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(reportSiteLocation):
        os.mkdir(reportSiteLocation)

def dtdd(dt, dd):
    return "{}\n:    {}\n\n".format(dt, dd)
    
"""
Hide need to include DATA_LOCN_TEMPL and SCHEMA
"""
def filteredResultIterator(stationNo, flTyp, useRF=False, startAtReply=""):
    locn = DATARF_LOCN_TEMPL.format(stationNo) if useRF else DATA_LOCN_TEMPL.format(stationNo)
    return FilteredResultIterator(locn, flTyp, startAtReply=startAtReply)
    
def loadSelectTypes(stationNo):
    return  json.load(open(SCHEMA_LOCN_TEMPL.format(stationNo) + "SELECT_TYPES.json"))

# ############################### Stat Utils #########################

"""
Shows off key stats with a little (obvious) stats observations
"""
def reportKeyStats(kstats):
    print("Count Values: {}".format(kstats["count"]))
    mn = kstats["mean"]
    print("Mean: {}".format(mn))
    M = kstats["median"]
    print("Median (M): {}".format(M))
    if M > mn:
        print("\t... M bigger than Mean so data skews left")
    elif M < mn:
        print("\t... M smaller than Mean so data skews right")
    print("Standard Deviation (s): {}".format(kstats["std"]))
    print("\treflect skew of min {} and max {}".format(kstats["min"], kstats["max"]))
    print("IQR: {}".format(kstats["iqr"]))                                                   
    print("\tbetween 1st ({}) and 3rd ({}) quartiles".format(kstats["quartileOne"], kstats["quartileThree"]))
    if "olt" in kstats:
        print("Outliers below", kstats["olt"])
    print("Outliers above\n", kstats["oht"])
    
"""
Key stats for distribution of values

https://docs.python.org/3/library/statistics.html

REM: values may be distributed ie/ not all responses the same size but are they
skewed too with a few exceptions off in a left/negative or right/positive tail? 

Note: ignoring sample (statistic) vs population (parameter). 

Basics which work for skewed distributions (typical):
... left (exceptions negative or left or smaller) or right where mean!=median
... median much better measure of centeredness as isn't effected by outliers
- min, max, mean
- quartile 1 2 (median), 3 ie/ three cut points (quintile is five)
  ... Medieval Latin quartilis from Classical Latin quartus, fourth
... "five-number summary" ideal for data w/ outliers and a 'box-and-whisker' plot 
... Quartiles divide a rank-ordered data set into four equal parts at 25, 50, 75
percentiles
- IQR ... "midspread", inter-quartile range
- olt < (Q1-1.5*IQR) ... 1.5 == "inner fences"
- oht > Q3+1.5*IQR ... 3 == "outer fences"

For not skewed (mean = median)
- standard deviation (using population form, not n-1 sample form)
- co-efficient of variation

Note: this is 'Univariate method' (See: https://www.kdnuggets.com/2017/01/3-methods-deal-outliers.html / 'Minkowski error' etc)

https://en.wikipedia.org/wiki/Outlier (Tukey's fences)
- above: Q3 + K(Q3-Q1) where K is 1.5 and K = 3 is "far out"
- below: Q1 - K(Q1-Q3)
ie/ K times Quartile. Fences for box plots.
"""    
def keyStats(values):
    try: # tmp allow for < 3.8
        # argument n=4 default => quartile from quantile method
        quartiles = statistics.quantiles(values, method="inclusive") # inc matched numpy
    except:
        import numpy
        quartiles = [numpy.percentile(values, 25), numpy.percentile(values, 50), numpy.percentile(values, 75)]
    quartileOne = quartiles[0] # numpy.percentile(values, 25)
    median = quartiles[1] # numpy.percentile(values, 50)
    quartileThree = quartiles[2] # numpy.percentile(values, 75)
    # get variance in middle 50% of data
    # numpy.subtract(*numpy.percentile(values, [75, 25]))
    iqr = quartileThree - quartileOne # interquartile range    
    oht = quartileThree + (1.5 * iqr) # Outlier High above this
    cntValsGTOHT = sum(1 for val in values if val > oht) 
    ohto = quartileThree + (3 * iqr) # idea of "outer fence": TODO: get name
    cntValsGTOHTO = sum(1 for val in values if val > ohto) 
    mx = sorted(values)[-1] # numpy.max(values)
    mn = sorted(values)[0] # numpy.min(values)
    meen = statistics.mean(values)
    pstd = statistics.pstdev(values) # for Population
    cvar = pstd/meen # co-efficient of variation
    stats = {"total": sum(val for val in values), "count": len(values), "min": mn, "quartileOne": quartileOne, "median": median, "quartileThree": quartileThree, "max": mx, "iqr": iqr, "oht": oht, "ohto": ohto, "mean": meen, "pstd": pstd, "cvar": cvar}

    olt = quartileOne - (1.5 * iqr) # Outlier Low below this
    if int(olt) > 0: # only olt if > 0 ie/ negative ie/ all acceptable
        stats["olt"] = olt
    """
    Own: should show outliers? only go into outliers if big ie/ if pretty smooth 
    """
    stats["hasOutliers"] = True if stats["max"] - stats["oht"] > stats["median"] else False
    if cntValsGTOHT:
        stats["gtoht"] = cntValsGTOHT
    if cntValsGTOHTO:
        stats["gtohto"] = cntValsGTOHTO
    return stats
    
"""
Use typically for byValueCount off typer or a Counter

    "a": 1
    "b": 2
    "c": 1
    
-----> [a, b, b, c]

and can pass into kstats if forceInt

Note: forceInt is for the measured values which may be integers but
are in string form.
"""
def flattenFrequencyDistribution(freqDistrib, forceInt=True):
    freqDistribFlat = []
    for entry in freqDistrib:
        f = [int(entry) if forceInt else entry for i in range(int(freqDistrib[entry]))]
        freqDistribFlat.extend(f)
    return freqDistribFlat
    
# ############################ Simple Report Utils #################

"""
If X.0 -> 0 otherwise to 2 pts (or whatever is passed in)
"""
def roundFloat(no, to=2):
    no = no.round(to)
    if no.is_integer():
        return int(no)
    return no
    
# ############################ Plot Utils ##########################

"""
Crude Example of Graph

import matplotlib.pyplot as plt

See: https://python-graph-gallery.com/120-line-chart-with-matplotlib/ and
https://matplotlib.org/gallery/index.html

    plot([2007, 2008, 2009, 2010], [1, 2, 3, 4])

"""
def plot(xs, ys):

    import matplotlib.pyplot as plt
    plt.plot(xs, ys, "ro")
    plt.xlabel("Time")
    plt.ylabel("Ibs")
    plt.show()
            
# ######################### Reduce Users ###########################

"""
Returns Users, Indexed 

Note: RF part ie/ 200 into DataRF will move out of here. Tmp index
maker for any type (including RF types) will become one simple utility.

TODO: go to lookup by IEN as simpler

Note: not dynamic if primary_menu_option or creator or title or # props >= 18 (or 16?)
"""
def reduce200(stationNo, onlyUsers=None, forceNew=False, coreOnly=False):

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "userReduction{}.json".format("CORE" if coreOnly else "")
    
    if not forceNew:
        try:
            infosByRef = json.load(open(jsnFl))
        except:
            pass
        else:
            return infosByRef

    allThere, details = checkDataPresent(stationNo, [
        # Additional settings/properties:
        # - duration (in seconds)
        # - force count on
        #     ipv4_address, remote_station_id, remote_user_ien, duration
        {"fileType": "200", "check": "ALL"} # Index
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))

    print("First time through, creating index by id for Users (200) for {}".format(stationNo))
    
    dataLocn = DATA_LOCN_TEMPL.format(stationNo)
    resourceIter = FilteredResultIterator(dataLocn, "200")
    infosByRef = {}
    cnt = 0
    for resource in resourceIter:
        cnt += 1
                
        if cnt % 1000 == 0:
            print("Processed another 1000 users - now at {:,}".format(len(infosByRef)))
        
        if onlyUsers and resource["_id"] not in onlyUsers:
            continue
        
        userId = "{} [{}]".format(resource["label"], resource["_id"])
                
        info = {"hasProps": sorted([key for key in resource if key not in set(["_id", "label", "type"])])}
        
        if "initial" in resource: # ex/ SMITH, JOE M --> JMS etc
            info["initial"] = resource["initial"]
        
        if "creator" in resource:
            # Only v1.1 will even have this ie/ not V10 on
            if resource["creator"]["id"] != "200-0":
                creatorRef = "{} [{}]".format(resource["creator"]["label"], resource["creator"]["id"])
                info["creator"] = creatorRef
        
        if "date_entered" in resource:
            info["date_entered"] = resource["date_entered"]["value"]
            
        if "title" in resource:
            info["title"] = resource["title"]["label"] # it's a pointer
            
        if "sex" in resource:
            info["sex"] = resource["sex"] # old fashioned gender
            
        if "ssn" in resource:
            info["ssn"] = resource["ssn"]
            
        # used for lookup BEFORE SSN in setup. goes with subject_organizations etc
        # ... comment says within VA, VA SSOi, secid is the unique field
        if "secid" in resource:
            info["secid"] = resource["secid"]
            # MOST are Dept of Vet Affairs - some VHA
            if "subject_organization" in resource:
                if resource["subject_organization"] not in ["Department Of Veterans Affairs", "Veterans Health Administration"]:
                    print("** WARNING: Expected Subject Org for sec id to always be VA! but {}".format(resource["subject_organization"]))
                if resource["subject_organization"] == "Veterans Health Administration":
                    info["is_subject_organization_vha"] = True
            
        # leave termination_reason for props list (ie/ presence or not)
        if "termination_date" in resource:
            info["termination_date"] = resource["termination_date"]["value"]
            
        if "network_username" in resource:
            info["network_username"] = resource["network_username"]   
            
        # adupn == va.gov email ... just check for now
        if "adupn" in resource and not re.search(r'va.gov$', resource["adupn"]):
            print("** unexpected adupn - email - not va.gov: {}".format(resource["adupn"]))
            if "email_address" in resource and resource["email_address"] != resource["adupn"]:
                print("** unexpected adupn - doesn't match email_address property: {}".format(resource["adupn"], resource["email_address"]))            
            
        """
        Latest entries can have unique_user_id BUT mainly == secid or is email.
        
        Issue with not taking as is maybe that intention to use for lookup may
        be missed - no as obvious there as listed in has_props
        """
        if "unique_user_id" in resource and not re.sub(r' ', '', resource["unique_user_id"]) == "":
            if re.search(r'\@', resource["unique_user_id"]):
                if "email_address" in resource and resource["unique_user_id"] != resource["email_address"]:
                    info["distinct_unique_user_id"] = resource["unique_user_id"]
            elif ("secid" in info and resource["unique_user_id"] != info["secid"]) or "secid" not in info:
                info["__distinct_unique_user_id"] = resource["unique_user_id"]
                                            
        if "last_signon_date_time" in resource:
            info["last_signon_date_time"] = resource["last_signon_date_time"]["value"]
            
        infosByRef[userId] = info 
            
        if coreOnly:
            continue
                            
        if "user_class" in resource:
            if len(resource["user_class"]) > 1:
                raise Exception("Expected only one user class assertion - ie/ singleton multiple")
            info["user_class"] = [uci["user_class"]["label"] for uci in resource["user_class"]][0]
            
        if "file_manager_access_code" in resource and resource["file_manager_access_code"] == "@":
            info["has_fileman_programmer_access"] = True
                
        if "primary_menu_option" in resource:
            info["primary_menu_option"] = resource["primary_menu_option"]["label"]
                
        if "secondary_menu_options" in resource:
            smos = set() # allows for dups with crude SMO addition
            for smo in resource["secondary_menu_options"]:
                if "secondary_menu_options" not in smo:
                    continue
                if not re.match(r'\d+$', smo["secondary_menu_options"]["label"]):
                    smos.add(smo["secondary_menu_options"]["label"])
            if len(smos):
                info["secondary_menu_options"] = sorted(list(smos))
            
        if "delegated_options" in resource:
            dos = set()
            for delo in resource["delegated_options"]:
                if "delegated_options" not in delo:
                    continue # seen in 663
                if not re.match(r'\d+$', delo["delegated_options"]["label"]):
                    dos.add(delo["delegated_options"]["label"])
            if len(dos):                   
                info["delegated_options"] = sorted(list(dos))
            
        if "keys" in resource:
            keys = set() 
            for key in resource["keys"]:
                if "key" not in key:
                    continue
                if not re.match(r'\d+$', key["key"]["label"]):
                    keys.add(key["key"]["label"])
            if len(keys):
                info["keys"] = sorted(list(keys))   
                
        if "preferred_editor" in resource: # for REMOTE user checks
            info["preferred_editor"] = resource["preferred_editor"]["label"]
            
        if "terminal_type_last_used" in resource:
            info["terminal_type_last_used"] = resource["terminal_type_last_used"]["label"] 
            
        # Consider "first sign on"
        if "visited_from" in resource:
            finalVisitedTime = None
            visitedFroms = {}
            for i, mresource in enumerate(resource["visited_from"]):
                if "visited_from" in mresource: 
                    visitedFrom = re.sub(r'[^\d]+', '', mresource["visited_from"])
                    duzAtHomeSite = mresource["duz_at_home_site"] if "duz_at_home_site" in mresource else ""
                    siteName = mresource["site_name"] if "site_name" in mresource else ""
                    siteId = ":".join([visitedFrom, duzAtHomeSite, siteName])
                    details = None
                    if "first_visit" in mresource or "last_visited" in mresource:
                        details = {}
                        if "first_visit" in mresource:
                            details["first"] = mresource["first_visit"]["value"]
                        if "last_visited" in mresource:
                            details["last"] = mresource["last_visited"]["value"]                            
                    visitedFroms[siteId] = details
            if len(visitedFroms):
                info["signed_on_froms"] = visitedFroms
            
    print("Serializing and Returning Index of {:,} Users (200)".format(cnt))
    if coreOnly:
        print("\t... core only")

    json.dump(infosByRef, open(jsnFl, "w"), indent=4)
    
    return infosByRef
    
# ############################ Reduce Patient ########################

"""
if "date_of_death" in info: # note: may or may not have date_entered_into_file too
    isDeceased
... 1/5 or more.
"""
def reduce2(stationNo, forceNew=False):

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "patientReduction.json"
    
    if not forceNew:
        try:
            infosByIEN = json.load(open(jsnFl))
        except:
            pass
        else:
            return infosByIEN

    print("First time through, creating patient entered reduction for {}".format(stationNo))
    sys.stdout.flush()
    
    dataLocn = DATA_LOCN_TEMPL.format(stationNo)
    resourceIter = FilteredResultIterator(dataLocn, "2")
    infosByIEN = {}
    PROPS_TO_TAKE = set([
        "name", 
        "sex", 
        "date_of_birth", 
        "social_security_number", 
        
        "who_entered_patient", 
        "date_entered_into_file", 
        
        "city", 
        "state", 
        "zip_code", 
        
        "period_of_service", 
        
        "date_of_death", 
        "death_entered_by", 
        
        "primary_eligibility_code", 
        "combat_service_indicated", 
        
        "preferred_facility", 
        
        # TODO: expand these - ensure all id's used by VA/C are in here
        "integration_control_number", 
        "subscription_control_number", # 774 entries for CIRN (see if sync file
        "full_icn",
        "primary_long_id", # see queue on where differs from SSN
        "primary_short_id" # may be redundant but get to see   
    ])
    """
    What should be
    - short_id == end of long_id
    - long_id == - version of ssn ie/ 111-22-3333 vs 111223333
    - icn is first part of full_icn
    ... ie/ nixing redundancies
    """
    def cutIds(info):
        if "integration_control_number" in info:
            if "full_icn" in info and info["integration_control_number"] == info["full_icn"].split("V")[0]:
                del info["integration_control_number"]
        if "primary_short_id" in info:
            if "primary_long_id" in info and info["primary_short_id"] == info["primary_long_id"].split("-")[-1]:
                del info["primary_short_id"]
        # in WWW, only if bug
        if "social_security_number" in info and "primary_long_id" in info:
            if info["social_security_number"] == re.sub(r'\-', "", info["primary_long_id"]):
                del info["primary_long_id"]
    for resource in resourceIter:
        patientIEN = resource["_id"].split("-")[1]
        info = {"ien": patientIEN}
        for prop in PROPS_TO_TAKE:
            if prop in resource:
                if isinstance(resource[prop], list):
                    raise Exception("Not reducing multiples of 2")
                if isinstance(resource[prop], dict):
                    if "value" in resource[prop]:
                        info[prop] = resource[prop]["value"]
                    else:
                        info[prop] = resource[prop]["label"]
                else: 
                    info[prop] = resource[prop]
        cutIds(info)
        infosByIEN[patientIEN] = info
        if len(infosByIEN) % 1000 == 0:
            print("Processed another 1000 patients - now at {:,}".format(len(infosByIEN)))
            sys.stdout.flush()
    
    print("Serializing and Returning Index of {:,} Patients (2)".format(len(infosByIEN)))
    sys.stdout.flush()

    json.dump(infosByIEN, open(jsnFl, "w"), indent=4)
    
    return infosByIEN
    
# ################# Reduce Location (44) n Institution (4) ####################

"""
Like other "anchor file" reductions, preludes to more formal reframes (RF) that
split out different types of key data into distinct types AND establishes a
lookup 

<----- will see if AppointmentHospitalLocation can be a sub class or split out appt info

Note: allows RF take - right now that RF is just the result of extracting the
44.001 and other multiples.
"""
def reduce44(stationNo, useRF=True): # do from RF

    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "hlReduction.json"
    try:
        infosByIEN = json.load(open(jsnFl))
    except:
        pass
    else:
        return infosByIEN
    print("First time through, creating hospital location reduction for {}".format(stationNo))
    sys.stdout.flush()
    infosByIEN = {}
    countWProvider = 0
    countDefaultProviderNotInProvider = 0
    def copyMProp(resource, red, mprop):
        if mprop not in resource:
            return None
        red[mprop] = set()
        for mresource in resource[mprop]:
            pref = "{} [{}]".format(mresource[mprop]["label"], mresource[mprop]["id"].split("-")[1])
            red[mprop].add(pref)
        red[mprop] = list(red[mprop])
        return red
    institutionsByHPrefix = defaultdict(lambda: Counter()) # for mapping
    cnt = 0
    for resource in filteredResultIterator(stationNo, "44", useRF=useRF):
        ien = resource["_id"].split("-")[1]
        red = {"ien": ien, "label": resource["name"]}
        infosByIEN[ien] = red
        cnt += 1
        if (cnt % 200) == 0:
            print("processed another 200 to {:,}".format(cnt))
            sys.stdout.flush()
        for litProp in ["abbreviation", "patient_friendly_name", "physical_location", "length_of_appointment", "default_appointment_type", "noncount_clinic_y_or_n", "clinic_meets_at_this_facility"]: # includes enum
            if litProp not in resource:
                continue
            red[litProp] = resource[litProp]
        for dtProp in ["inactivate_date", "reactivate_date", "availability_flag"]:
            if dtProp not in resource:
                continue
            red[dtProp] = resource[dtProp]["value"]
        for refProp in ["stop_code_number", "credit_stop_code", "institution", "division", "type_extension", "default_provider"]:
            if refProp not in resource:
                continue
            refv = resource[refProp]
            ref = "{} [{}]".format(refv["label"], refv["id"].split("-")[1])
            red[refProp] = ref
            if refProp == "institution":
                hprefix = re.sub(r'^ZZ', '', re.split(r'[ \-\/]', resource["name"])[0])
                if len(hprefix):
                    institutionsByHPrefix[hprefix][ref] += 1
        # Mult Availability - note many exs where no appts
        if "availability" in resource: # want to see if always there if appointments too
            red["has_availability"] = True
        if copyMProp(resource, red, "provider"):
            if "default_provider" in red and red["default_provider"] not in red["provider"]:
                countDefaultProviderNotInProvider += 1
            countWProvider += 1
        copyMProp(resource, red, "privileged_user") # the clerks etc
        # Note: appointment FLIPPED/REDUCED so won't know that?
    
    """    
    Fill in missing instit links based on prefix mappings from those with an instit
    
    Note that based on error'ed assignment of prefixes, > 1 institution my be assigned
    a prefix. We will take majority rule so it there is > one then 
    """
    ninstitutionsByHPrefix = {}
    for hprefix in institutionsByHPrefix:
        if len(list(institutionsByHPrefix)) == 1:
            ninstitutionsByHPrefix[hprefix] = list(institutionsByHPrefix)[0]
            continue
        useTotal = sum(institutionsByHPrefix[hprefix][instit] for instit in institutionsByHPrefix[hprefix])
        for instit in institutionsByHPrefix[hprefix]:
            if float(institutionsByHPrefix[hprefix][instit])/useTotal >= 0.75:
                ninstitutionsByHPrefix[hprefix] = instit
                break
    noMissing = 0
    noFilled = 0
    for ien in infosByIEN:
        info = infosByIEN[ien]
        if "institution" in info:
            continue
        noMissing += 1
        hprefix = re.split(r'[ \-\/]', info["label"])[0]
        if hprefix in ninstitutionsByHPrefix:
            info["institution"] = ninstitutionsByHPrefix[hprefix]
            noFilled += 1
    print("Of {:,} locations missing an institution link, {:,} were filled by heuristic".format(noMissing, noFilled))
        
    print("Dumping {:,} locations, {:,} have providers with {:,} having default provider not in provider list".format(len(infosByIEN), countWProvider, countDefaultProviderNotInProvider))
    json.dump(infosByIEN, open(jsnFl, "w"), indent=4)
    return infosByIEN

"""
name, short_name | names
status, inactive_facility_flag | National or Local and if active. Nationals are maintained by MFS and should have a station number (expect Nationals same across VistAs)
station_number, official_va_name | Nationals should have these and they should not be changed by a site (expect same cross VistAs)
facility_type | facility_type (pter to meta file) replaces va_type_code
reporting_station | "pointer back to the Institution file that indicates the site that does the reporting to VACO." TODO: if SPOK for clinics etc
location_timezone | time zone of location
multi-division facility | TODO: see if useful

For now ignoring, address, agency_code etc

TODO: push into reduction useful for nationals -- use association
"""
def reduce4(stationNo):
    jsnFl = TMPWORKING_LOCN_TEMPL.format(stationNo) + "4Reduction.json"
    try:
        infosByIEN = json.load(open(jsnFl))
    except:
        pass
    else:
        return infosByIEN
    print("First time through, creating institution reduction for {}".format(stationNo))
    infosByIEN = {}
    for resource in filteredResultIterator(stationNo, "4"):
        ien = resource["_id"].split("-")[1]
        red = {"ien": ien, "label": resource["name"]}
        infosByIEN[ien] = red
        # OPEN QUESTION: any useful or now used for national ... see
        for refProp in ["reporting_station", "facility_type", "location_timezone"]:
            if refProp not in resource:
                continue
            red[refProp] = "{} [{}]".format(resource[refProp]["label"], resource[refProp]["id"].split("-")[1])
        if "inactive_facility_flag" in resource:
            red["isInactive"] = True
        if "status" in resource:
            if resource["status"] == "N:National":
                red["isNational"] = True
        for prop in ["short_name", "multi_division_facility", "npi", "official_va_name", "station_number"]:
            if prop not in resource:
                continue
            red[prop] = resource[prop]
        # VISN, Parent Facility etc
        if "associations" in resource:
            for association in resource["associations"]:
                if "parent_of_association" not in association:
                    continue
                atyp = association["associations"]
                apred = re.sub(r' ', '_', atyp["label"].lower())
                ref = association["parent_of_association"]
                if ref["id"] == resource["_id"]:
                    continue # avoid self reference in Puget being parent of Puget
                red[apred] = ref
                if apred == "VISN":
                    try:
                        visnNo = int(ref["label"].split(" ")[1])
                    except:
                        raise Exception(f'Institution {red["ien"]}/{red["label"]} has non numeric VISN value {ref["label"].split(" ")[1]}')
                    res["visn"] = visnNo   
        # EX/ VISTANUM == stationNumber but npi too (also explicit?)
        # ... has status and effective_date_time too
        # ... https://npiregistry.cms.hhs.gov/registry/provider-view/1477502946 ... 
        # America Lake
        if "identifier" in resource:
            for identifier in resource["identifier"]:
                ityp = identifier["coding_system"]
                ipred = re.sub(r' ', '_', ityp.lower()) + "_id"
                id02 = identifier["id__02"]
                red[ipred] = id02
        # taxonomy_code - reduce to list
    json.dump(infosByIEN, open(jsnFl, "w"), indent=4)
    return infosByIEN
    
"""
Utility for making tmp's - flatten out dates, pters etc

TODO: add to reduction utilities.
"""
def flattenPropValues(resource, nixTypeLabel=True): # not recursive
    if nixTypeLabel:
        for prop in ["type", "label"]:
            if prop in resource:
                del resource[prop]
    for prop in resource.keys():
        if isinstance(resource[prop], list):
            for mresource in resource[prop]:
                flattenPropValues(mresource, False)
            continue
        if not isinstance(resource[prop], dict):
            continue
        if "value" in resource[prop]:
            resource[prop] = resource[prop]["value"]
            continue
        if "id" in resource[prop]:
            resource[prop] = "{} [{}]".format(resource[prop]["label"], resource[prop]["id"])
            
# ##################### Generate/Use VISN/Station Institution Map ###################

from .visnNVistAs import byVISNByVistAStation

"""
To use:
   > vistasOfVISNByOne("668")
"""
def vistasOfVISNByOne(oneVistASNO):
    return vistasOfVISN(visnOfVistA(oneVistASNO))

"""
To use:
   > vistasOfVISN("20") -- all VISN 20 VistAs as {"{SNO}": "{MN}" ...}
"""
def vistasOfVISN(onlyVISN=""):
    mainsByVISN = defaultdict(list)
    for visn in sorted(byVISNByVistAStation, key=lambda x: int(x)):
        for sno in byVISNByVistAStation[visn]:
            mainInstance = byVISNByVistAStation[visn][sno][0]
            if "mn" not in mainInstance:
                continue
            mainsByVISN[visn].append({sno: mainInstance["mn"]})
    if onlyVISN == "":
        return mainsByVISN
    return mainsByVISN[onlyVISN]

"""
To use:
   > visnOfVistA("668") -- get "20"
"""
def visnOfVistA(stationNo):
    for visn in byVISNByVistAStation:
        if stationNo in byVISNByVistAStation[visn]:
            return visn
    return ""

"""
Different from reduce4, this produces the VISN/VistA-based Station Map which goes into
fmqlreports itself at the top level.

Default saved to tmp of specific vista but main use is to populate/refresh the visn/vista
definition of visnNVistAs.py in fmqlreports.

20: {
    668: [
        { spokane
        { coeur ...
}

Revisit: may remove veterans homes etc (don't appear to have NPIs)

TODO: 
- add in geo data based on addresses and urls and reproduce something like
(but better than) https://www.va.gov/directory/guide/region.asp?ID=1010 
- find how to add mn's to agreed values (ex/ WCO not WCI for White City). Not in 4? so fixing for VISN 20 but beyond?
"""
def reduce4VISNVistASNO(stationNo, saveTo=""):

    # Pass 1: take relevant institutions (remaining outliers deleted in phase 2)
    dataLocn = f'{VISTA_DATA_BASE_DIR}{stationNo}/Data'
    resourceIter = FilteredResultIterator(dataLocn, "4") 
    stationNosSeen = set()
    instits = []
    for resource in resourceIter:
        if not (
            "status" in resource and 
            resource["status"] == "N:National"
        ):
            continue
        if "inactive_facility_flag" in resource:
            continue
        if "station_number" not in resource:
            continue
        if "associations" not in resource:
            continue
        # There are some CEMETERY before but its a mix
        if int(resource["station_number"][0:3]) >= 780:
            continue 
        if int(resource["station_number"][0:3]) < 400:
            continue 
        # Only do top level with locked IENs
        if len(resource["station_number"]) == 3 and resource["station_number"] != resource["_id"].split("-")[1]:
                continue
        if resource["station_number"] in stationNosSeen:
            raise Exception("Station No reused")
        stationNosSeen.add(resource["station_number"])
        instit = {"name": resource["name"], "stationNumber": resource["station_number"]}
        if "official_va_name" in resource and resource["official_va_name"] != resource["name"]:
            instit["officialName"] = resource["official_va_name"]
        if "npi" in resource:
            instit["npi"] = resource["npi"]
        if "domain" in resource:
            instit["domain"] = resource["domain"]["label"]     
        visnAssocs = [assoc for assoc in resource["associations"] if "associations" in assoc and assoc["associations"]["label"] == "VISN" and "parent_of_association" in assoc and assoc["parent_of_association"]["label"].split(" ")[0] == "VISN"]
        if len(visnAssocs) > 1:
            raise Exception("Ambiguous VISN declaration")
        if len(visnAssocs): # missing filled in below
            instit["visn"] = visnAssocs[0]["parent_of_association"]["label"].split(" ")[1]
        instits.append(instit)
        
    # Pass 2: 3 purposes - [1] nix entries with 'bad' stations ex/ 452XX that don't
    # fit into any VistA, [2] puts main center of a VistA first AND [3] gives normalized
    # artificial IEN to entries ie/ global IENs
    stationGroups = dict((instit["stationNumber"], [instit]) for instit in instits if len(instit["stationNumber"]) == 3)
    visnStationGroups = defaultdict(dict)
    MNFIXED = {"363": "ANC", "531": "BOI", "687": "WWW", "668": "SPO", "648": "POR", "653": "ROS", "663": "PUG", "692": "WCO", "757": "COS"} # 20 + Columbus
    for sgno in stationGroups:
        grpVISNs = set(instit["visn"] for instit in stationGroups[sgno] if "visn" in instit)
        if len(grpVISNs) != 1:
            raise Exception("A Vista/Station Group has no VISN or > 1 VISN")
        mainInstit = stationGroups[sgno][0]
        mainInstit["ien"] = sgno # ie/ 668 IEN is 668
        if "visn" not in mainInstit:
            raise Exception("Expect all main hospitals of VISN to spec VISN")
        visnStationGroups[mainInstit["visn"]][sgno] = [mainInstit]
        if sgno in MNFIXED:
            mainInstit["mn"] = MNFIXED[sgno]
            continue
        if "domain" in mainInstit:
            mainInstit["mn"] = re.sub(r'\-', '', mainInstit["domain"])[0:3]
    for instit in sorted(instits, key=lambda x: x["stationNumber"]):
        if len(instit["stationNumber"]) == 3:
            continue
        sgno = instit["stationNumber"][0:3]
        if sgno not in stationGroups: # Purge extras - not for VistA
            continue
        stationGroups[sgno].append(instit)
        # Ex/ 687HA would be 68701, 68702 ... ie/ up to 99 places and 687
        # ie/ three long is for main center as it is now.
        instit["ien"] = int("{}{}".format(sgno, str(len(stationGroups[sgno]) - 1).zfill(2)))         
        if "visn" not in instit:
            instit["visn"] = stationGroups[sgno][0]["visn"] 
        visnStationGroups[instit["visn"]][sgno].append(instit)  
        
    red = {
        "about": {
            "generated": datetime.now().strftime("%Y-%m-%d"), 
            "fromVistA": stationNo, 
            "fromFile": 4
        },
        "byVISNByVistAStation": visnStationGroups
    }
    jsnFl = saveTo if saveTo else TMPWORKING_LOCN_TEMPL.format(stationNo) + "4VISNVistASNOReduction.json"
    json.dump(red, open(jsnFl, "w"), indent=4)
    return red 
    
# ############################ Simple Reusable MD Makers ###############################

def mdFilesOfDomain(stationNo, domainName, bottomFileNo, beyondFileNo, usedFiles):

    selectTypes = loadSelectTypes(stationNo)

    fileTypeInfos = [typeInfo for typeInfo in selectTypes["results"] if "number" in typeInfo and float(typeInfo["number"]) >= bottomFileNo and float(typeInfo["number"]) < beyondFileNo and "count" in typeInfo]
     
    mu = """There are <span class="yellowIt">{:,}</span> files in the system dedicated to the _{}_ ...
    
""".format(len(fileTypeInfos), domainName)
    
    tbl = MarkdownTable(["Name", "\#", "Count"])
    for typeInfo in fileTypeInfos:
        tbl.addRow([
                "__{}__".format(typeInfo["name"]) if typeInfo["name"] in usedFiles else typeInfo["name"], 
            typeInfo["number"], 
            typeInfo["count"]
    ])
    mu += tbl.md() + "\n\n"
        
    return mu
    
# ##################### Plot Creation #########################
        
"""
Plot creation is separate --- may move into Visualize so common

ex/ plotDataFile = "plotDataIFC.json" ie/ plotDataFileSuffix="IFC"

TODO:
- ifcPlTravelVISNSankey (from Travel PlacersVISNSankey) -- .md change too
- all consult services categ ... -> ifcAllConsultsCategory 
- remove 0 ie/ NO OTHER!
- map SEA -> PUG, MAN -> SPO

"""

def makePlots(stationNo, plotDataFileSuffix):   

    import io
    import pandas as pd

    from visualize import Visualize
    imageDir = f'{VISTA_DATA_BASE_DIR}{stationNo}/ReportSite/Images'
    if not os.path.isdir(imageDir):
        os.mkdir(imageDir)
    viz = Visualize(imageDir)
    plotData =  json.load(open(f'{VISTA_DATA_BASE_DIR}{stationNo}/TmpWorking/plotData{plotDataFileSuffix}.json'))
    for plotDataName in plotData:
        plotDataInfo = plotData[plotDataName]
        print(f'Remaking {plotDataInfo["title"]}')
        method = getattr(viz, plotDataInfo["plotMethod"])
        if "rows" in plotDataInfo:
            arg = viz.makeDF(
                data=plotDataInfo["data"],
                columns=plotDataInfo["columns"],
                rows=plotDataInfo["rows"]
            )
        elif "specs" in plotDataInfo:
            arg = plotDataInfo["specs"]
        elif "dfcsv" in plotDataInfo:
            dfcsv = plotDataInfo["dfcsv"]
            df = pd.read_csv(
                io.StringIO("\n".join(dfcsv)),
                index_col=0, 
                parse_dates=True
            )
            arg = df
        else:
            raise Exception("Expect plot data for DF or as 'specs'")
        kargs = plotDataInfo.get("kargs", {})
        plotRef = method(arg, plotDataInfo["title"], plotDataInfo["plotName"], **kargs)
        print(f'\texpect reference to {plotRef} in report')
        
def muPlotRef(plotInfo, flushToDirectory):
    """
    # Want separate from visualize for now BUT possible break
    # as separate from Visualize naming impl
    
    ex/ ![ifcTravelVISNSankey](Images/ifcTravelVISNSankeyBGW.svg)
    """
    def svgPlotFile(plotName, plotMethodName, plotKArgs, flushToDirectory):
        """
        So don't have to make the plot but can know its name as making md
        Compatible with Visualize.__flushPlot's naming of SVG's
        """
        coreMethodName = re.sub("plot", "", plotMethodName)
        coreMethodNameU = coreMethodName[0].upper() + coreMethodName[1:]
        plotFilePrefix = f'{plotName}{coreMethodNameU}'
        if plotKArgs:
            if "start" in plotKArgs:
                plotFilePrefix += plotKArgs["start"]
            if "end" in plotKArgs:
                plotFilePrefix += "-{}".format(plotKArgs["end"])
        plotFile = f'{flushToDirectory}/{plotFilePrefix}BGW.svg'
        return plotFile, plotFilePrefix
    plotFile, plotFilePrefix = svgPlotFile(plotInfo["plotName"], plotInfo["plotMethod"], plotInfo["kargs"] if "kargs" in plotInfo else None, flushToDirectory)
    return "![{}]({})".format(plotFilePrefix, plotFile)
    
# ######################### Simple Type Utils ########################

class PropertyCounter:
    
    def __init__(self):
        self.__counts = Counter()
        self.__countResources = 0
            
    def resourceProperties(self, props):
        for prop in props:
            self.__counts[prop] += 1
        self.__countResources += 1
            
    def mandatories(self):
        return sorted([prop for prop in self.__counts if self.__counts[prop] == self.__countResources]) 
            
    def optionalCounts(self):
        return dict((prop, self.__counts[prop]) for prop in self.__counts if self.__counts[prop] < self.__countResources) 



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os

from fmqlutils.cacher.cacherUtils import flip

"""
Peer of flip44
"""
def flipAppointments(stationNo):
    from fmqlutils.cacher.cacherUtils import flip
    flip(stationNo, "2", "appointment", "2_98", "PATIENT APPOINTMENT")

# ################################# DRIVER #######################

def main():

    assert(sys.version_info >= (2,7))

    if len(sys.argv) < 2:
        print("need to specify station # ex/ 442 - exiting")
        return

    stationNo = sys.argv[1]

    flipAppointments(stationNo)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os

from fmqlutils.cacher.cacherUtils import flip

# flip(stationNo, "2", "appointment", "2_98", "PATIENT APPOINTMENT")

"""
A "multiple flipper" - RF of multiples ie/ simplest RF

Covers:
- 44_00[13] for Legacy Appointments out of 44
"""
def flipAppointments(stationNo):
    flip(stationNo, "44", "appointment", "44_001", "APPOINTMENT")
    # Note: leaving container prop == hospital location as is so pulled
    # in explicitly. 
    flip(stationNo, "44_001", "patient", "44_003", "PATIENT APPOINTMENT", explicitInheritProps=["appointment_date_time"], useRF=True)

# ################################# DRIVER #######################

def main():

    assert sys.version_info >= (3, 4)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    flipAppointments(stationNo)

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict
from datetime import datetime

from fmqlutils.typer.reduceTypeUtils import singleValue

"""
TODO:
- quick tighten: remote should all have key division; locals from non local divisions (what are they?)
- pure remotes (from all remotes)
- pure locals
- mixed ... of a type [may put locals back into local and give as remote older]

TOADD (beyond SO, 200)
- ImagingUsers(stationNo, _2006_82WSAllTypeInfo, _2006_82WSSubTypeDatas)
- 38.1 access too
"""

"""
Statically Classify Active (with SO) Users

(vs Dynamically if sniffing traffic and needs table showing dynamic equivalent if any of way it classifies ex/ IP is common or WS = 10 or remote app spec'ed => BSE etc)

Uses User (200), 3.081 (last year or some period up to now) reduction
"""
class UserClassifier:

    def __init__(self, stationNo, userInfoByUserRef, all3_081, st3_081sByUserRef, smallSO=False):
    
        if not (len(userInfoByUserRef) and len(st3_081sByUserRef)):
            raise Exception("Can't classify without both user infos AND SO types per user")
                
        if "_createDateProp" not in all3_081:
            raise Exception("Classified 3.081's must have create date prop")
        createDatePropInfo = all3_081[all3_081["_createDateProp"]]
        firstSODay = createDatePropInfo["firstCreateDate"].split("T")[0]
        lastSODay = createDatePropInfo["lastCreateDate"].split("T")[0]
        
        self.__result = { 
        
            "firstSODay": firstSODay,
            "lastSODay": lastSODay,
            
            "inactiveUserCount": 0,
            "remoteExcludeReasonCount": defaultdict(int),
                    
            "activeUserRefs": set(st3_081sByUserRef),
            
            "dodUserRef": "", # special machine 'proxy' (CAPRI, not av)
            "postmasterUserRef": "",
            
            "activeProxyUserRefs": set(), # machine proxy with av
            "activeNonProxyMachineUserRefs": set(), # machine
            "activeRemoteUserRefs": set(), # real people
            "activeRemoteUserRefsSubs": defaultdict(lambda: set()),
            "activeLocalUserRefs": set(), # real people
            
            "activeNotCategorizedUserRefs": set(),
            
            "warningsByUserRef": defaultdict(list)
                                                    
        }
                        
        # Make a reduced form of signed_on_froms for SO period if SO period long enough
        # ... may nix as not used much below any more (SO is master)
        def makeSignedOnFromsSO(overallFirstSODay, userInfoByUserRef):
            overallFirstSODayDT = datetime.strptime(overallFirstSODay, "%Y-%m-%d")
            deletedAsNone = 0
            for userRef in userInfoByUserRef:
                userInfo = userInfoByUserRef[userRef]
                if "signed_on_froms" not in userInfo:
                    continue
                signed_on_froms_so = {}
                for siteId in userInfo["signed_on_froms"]:
                    detail = userInfo["signed_on_froms"][siteId]
                    if detail == None or "last" not in detail:
                        continue
                    lastSODayDT = datetime.strptime(detail["last"].split("T")[0], "%Y-%m-%d")
                    if lastSODayDT < overallFirstSODayDT: 
                        continue
                    signed_on_froms_so[siteId] = detail
                if len(signed_on_froms_so):
                    userInfo["signed_on_froms"] = signed_on_froms_so
                else:
                    del userInfo["signed_on_froms"]
        if smallSO == False: # ie/ do a YR1 cut down
            makeSignedOnFromsSO(firstSODay, userInfoByUserRef)
        
        self.__stationNo = stationNo
        self.__userInfoByUserRef = userInfoByUserRef
        self.__st3_081sByUserRef = st3_081sByUserRef
                
    def classify(self):
    
        print("Classifying {:,} active (with SO) users of out {:,} total".format(
            len(self.__result["activeUserRefs"]), 
            len(self.__userInfoByUserRef)
        ))
        
        self.__result["postmasterUserRef"] = self.__enforcePostmaster() 
        self.__result["dodUserRef"] = self.__findDoDUser() # one off
                
        for userRef in self.__userInfoByUserRef:
        
            if userRef not in self.__result["activeUserRefs"]:
                self.__result["inactiveUserCount"] += 1
                continue
                
            if userRef in [self.__result["dodUserRef"], self.__result["postmasterUserRef"]]:
                continue
                
            userInfo = self.__userInfoByUserRef[userRef]
            signOnType = self.__st3_081sByUserRef[userRef]
            
            if self.__testProxyUser(userRef, userInfo, signOnType):
                self.__result["activeProxyUserRefs"].add(userRef)
                continue
            if self.__testNonProxyMachineUser(userRef, userInfo, signOnType):
                self.__result["activeNonProxyMachineUserRefs"].add(userRef)
                continue
            remoteUserSubCatag = self.__testRemoteUser(userRef, userInfo, signOnType)
            if remoteUserSubCatag:
                self.__result["activeRemoteUserRefs"].add(userRef)
                self.__result["activeRemoteUserRefsSubs"][remoteUserSubCatag].add(userRef)
                continue
            if self.__testLocalUser(userRef, userInfo, signOnType):
                self.__result["activeLocalUserRefs"].add(userRef)
                continue
                            
        print("Finished Classification")
                
        return self.__result
        
    """
    Postmaster
    - .5 IEN
    - pmo, smos, keys
    - signins
    
    Expect:
    - NO ssn
    """
    def __enforcePostmaster(self):
        postmasters = [userRef for userRef in self.__userInfoByUserRef if re.match(r'POSTMASTER \[', userRef)]
        if len(postmasters) == 0:
            raise Exception("No POSTMASTER")
        if len(postmasters) > 1:
            warning = "> 1 User Named Postmaster: {}".format("/".join(postmasters))
            print("** Warning PostMaster: {}".format(warning))
        postmaster = [userRef for userRef in postmasters if re.search(r'200\-\.5\]', userRef)]
        if len(postmaster) != 1:
            raise Exception("POSTMASTER not 200-.5")
        userRef = postmaster[0]
        if not re.search(r'\[200-\.5\]$', userRef):
            raise Exception("IEN not .5")
        if userRef not in self.__st3_081sByUserRef:
            raise Exception("No sign ins for POSTMASTER")
        userInfo = self.__userInfoByUserRef[userRef]
        if not ("primary_menu_option" in userInfo and "secondary_menu_options" in userInfo and "keys" in userInfo):
            warning = "Some of primary_menu_option, secondary_menu_options, keys missing"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning PostMaster: {}".format(warning))
        if "ssn" in userInfo:
            warning = "SSN not expected"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning PostMaster: {}".format(warning))
        if "date_entered" not in userInfo:
            warning = "No User 'date entered'"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning PostMaster: {}".format(warning))
        return userRef
        
    """
    Special Non Proxy Machine User for DoD and other remote accesses via CAPRI
    
        Most visited_froms in System AND
        No User Class AND 
        has SSN AND 
        has at least CAPRI, CPRS, VPR SMOs
        has sign ins
        
    Expect (warn otherwise):
        visited from 524:14358 (Braintree) and 506:10000029117 AND
        SSN is machine form AND
        signs are all LOA == 1 AND no remote_app setting (CAPRI)
        
    Not enforcing X:1 for station:IEN for most and then JLV IP/DoD user (may do
    later)

    Note: 663 shows alias 'VISITOR'
    """
    def __findDoDUser(self):
        candidatesByVisitedFromCount = defaultdict(list)
        for userRef in self.__userInfoByUserRef:
        
            userInfo = self.__userInfoByUserRef[userRef]
            
            if "user_class" in userInfo:
                continue
            if "ssn" not in userInfo:
                continue
            if "signed_on_froms" not in userInfo:
                continue
            if "secondary_menu_options" not in userInfo or len(set(["DVBA CAPRI GUI", "OR CPRS GUI CHART", "VPR APPLICATION PROXY"]) - set(userInfo["secondary_menu_options"])):
                continue
                
            candidatesByVisitedFromCount[len(userInfo["signed_on_froms"])].append(userRef)
            
        if len(candidatesByVisitedFromCount) == 0:
            raise Exception("Can't reduce DOD User - none!")            
        maxCount = sorted(candidatesByVisitedFromCount, reverse=True)[0]
        if len(candidatesByVisitedFromCount[maxCount]) > 1:
            raise Exception("Can't reduce DOD User - ambiguous!")
            
        userRef = candidatesByVisitedFromCount[maxCount][0]
        
        if userRef not in self.__st3_081sByUserRef:
            raise Exception("Expect DoD User to have sign ins")
        
        userInfo = self.__userInfoByUserRef[userRef]
        if sum(1 for sof in userInfo["signed_on_froms"] if re.match(r'523:14358', sof)) == 0:
            warning = "No 523:14358"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
        if sum(1 for sof in userInfo["signed_on_froms"] if re.match(r'506:10000029117', sof)) == 0:
            warning = "No 506:10000029117"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
        if not couldBeMachineSSN(userInfo["ssn"]):
            warning = "Machine SSN but {}".format(userInfo["ssn"])
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
            
        if "date_entered" not in userInfo:
            warning = "No User 'date entered'"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
            
        # Could add and MDWS remote_app as now set as default (used to be blank)
        signOnType = self.__st3_081sByUserRef[userRef]
        if not enforceLOA(signOnType, "1"):
            warning = "SO LOA 1 == CAPRI login"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
            
        if "device" not in signOnType or sum(1 for device in signOnType["device"]["byValueCount"] if not re.match(r'0', device)):
            warning = "Not All Zero Device"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
            
        # May be too aggressive but ok for now
        mds = manyDivisionSummary(signOnType)
        if mds:
            warning = "Not Default Division Only: {}".format(mds)
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning DoD User: {}".format(warning))
                        
        return userRef
                
    """
    Proxy User
        user_class CONNECTOR PROXY and 
        NO visited_from and
        access, verify and
        SO's all LOA 2 and
        NO remote_*'s set in SO <---- triple AV/LOA2/no remote_... == CPRS like
        
    Expect:
        if SSN then Machine SSN
        
    TODO:
    - workstation_name == TCP Connect patterns
    """
    def __testProxyUser(self, userRef, userInfo, signOnType):
    
        if "user_class" not in userInfo:
            return False
            
        if not re.search(r'PROXY', userInfo["user_class"]):
            return False
            
        if "signed_on_froms" in userInfo:
            raise Exception("Didn't expect proxy user {} to have visited_from".format(userRef))
            
        if len(set(["access_code", "verify_code"]) - set(userInfo["hasProps"])):
            raise Exception("Expected proxy user {} to have access verify".format(userRef))
                        
        if sum(1 for prop in signOnType if re.match(r'remote', prop)):
            # saw in BOI 531 and TERMINAL EMULATOR remote_app (may be overload use?)
            warning = "Unexpected \"remote_...\" property in SO of Proxy User"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef))
                                    
        # HMP SPOK only => need to show
        if userInfo["user_class"].split(" ")[0] != "CONNECTOR":
            warning = "Only CONNECTOR Proxies but {}".format(userInfo["user_class"].split(" ")[0])
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef))
            
        # Ignoring extras like "title" or ...
        unexpectedProps = [prop for prop in ["disuser", "disuser_date", "ssn"] if prop in userInfo["hasProps"]]
        if len(unexpectedProps):
            warning = "Inappropriate Prop(s) \"{}\"".format("/".join(unexpectedProps))
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef)) 
            
        if "date_entered" not in userInfo:
            warning = "No User 'date entered'"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef))           
            
        if not enforceLOA(signOnType, "2"):
            warning = "SO LOA not only 2"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef))
            
        mds = manyDivisionSummary(signOnType)
        if mds:
            warning = "Not Default Division Only: {}".format(mds)
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef))
            
        if "device" not in signOnType or sum(1 for device in signOnType["device"]["byValueCount"] if not re.match(r'0', device)):
            warning = "Not All Zero Device"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Proxy: {} for {}".format(warning, userRef))
                        
        return True

    """
    Not Proxy and Going off Keywords - does NOT include DoD User irrespective of name
    and POSTMASTER isn't here either.
    
    Expect:
    - machine SSN
    - visited_from 
    
    Note:
    - not enforcing LOA 2 as see 200, 2001 combos where first is 1 and then move to 2
    """
    def __testNonProxyMachineUser(self, userRef, userInfo, signOnType):
        
        if "user_class" in userInfo:
            return False
            
        if not re.search(r'(\,USER$|^USER\,|CVIX|CHCS|WEB ITS|RADIOLOGY|PHARMACY)', userRef):
            return False
                       
        missingExpectedProps = [prop for prop in ["secondary_menu_options", "ssn", "visited_from"] if prop not in userInfo["hasProps"]]
        if len(missingExpectedProps):
            warning = "Missing Prop(s) \"{}\"".format("/".join(missingExpectedProps))
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef)) 
           
        if "ssn" in userInfo and not couldBeMachineSSN(userInfo["ssn"]):
            warning = "Not Machine SSN"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))
                       
        # Ignoring extras like "title" or ...
        unexpectedProps = [prop for prop in ["disuser", "disuser_date", "primary_menu_option", "keys", "sex"] if prop in userInfo["hasProps"]]
        if len(unexpectedProps):
            warning = "Inappropriate Prop(s) \"{}\"".format("/".join(unexpectedProps))
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))  
            
        if "remote_app" in signOnType:
            warning = "\"remote_app\" in (some) SO"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))
            
        if "date_entered" not in userInfo:
            warning = "No User 'date entered'"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))
            
        if not enforceLOA(signOnType, "2"):
            warning = "SO LOA not only 2"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))            
            
        mds = manyDivisionSummary(signOnType)
        if mds:
            warning = "Not Default Division Only: {}".format(mds)
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))
            
        if "device" not in signOnType or sum(1 for device in signOnType["device"]["byValueCount"] if not re.match(r'0', device)):
            warning = "Not All Zero Device"
            self.__result["warningsByUserRef"][userRef].append(warning)
            print("** Warning Non Proxy: {} for {}".format(warning, userRef))
        
        return True
        

    PUREZEROSOCREATEDUSER_200_MAND_PROPS = set([
        "name", "alias", "name_components", "signature_block_printed_name", "date_entered", "xus_logon_attempt_count", "xus_active_user", "secondary_menu_options", "timestamp", "visited_from"
    ])    
    
    PUREZEROSOCREATEDUSER_200_OPT_PROPS = set([    

        # not in 668
        "initial",

        # nearly always EXCEPT never for Manilla
        "ssn",

        # Optional id's
        "network_username", "email_address",   

        # Never 442, Always 640, 668 - "SCREEN EDITOR - VA FILEMAN"
        "preferred_editor",     

        # Doesn't seem to effect CAPRI at all (in MM in some systems)
        "disuser", "entry_last_edit_date", "termination_reason",

        # Termination Date: This is the date after which the computer will no longer recognize this user's ACCESS CODE. Once this date has passed, when the USER TERMINATE job runs it will clean out this users data based on flags in the NEW PERSON file.
        # 640 has for them all ie/ sets when WILL term and then 
        # actual term kicks in
        "termination_date", # 640 and JLV user
        
        "last_signon_date_time" # MAND EXCEPT FOR VCB so putting here
    ]) 
    
    """
    Remote means comes from other station AND created by such sign ons (ie/ no clerk
    creates)
    
        Not Machine User (Proxy or Otherwise) AND
        0 Created AND
        Has only ALLOWED PROPERTIES AND
        Has All Mandatory Properties* AND
        if no SSN then Manilla
        signed on from AND
        signed on from at least one remote station
        
    ie/ ?? assuming only allowed props => no local logons
        
    MORE TODO: sub again
    - VISTAWEBPROD_IP = '10.224.74.88'
    - no SMO?
    - force ALL remotes and put others into "mixed"? <---- ****
    - WARN if REMOTEs have LOA != 2 ie/ BSE? and if remote_app == 100% sign ons
        
    """
    def __testRemoteUser(self, userRef, userInfo, signOnType):
    
        """
        Pure JLV Subset

        Since 2020, JLV WEB SERVICES as Context (only one) in JLV created users
        """
        def __testPureJLVSubset(userInfo, signOnType):
        
            # if remote app - must be MEDICAL DOMAIN WEB SERVICES
            if "remote_app" in signOnType and not (
                len(signOnType["remote_app"]["byValueCount"]) == 1 and 
                singleValue(signOnType, "remote_app").split(" [")[0] == "MEDICAL DOMAIN WEB SERVICES"):
                return False
                
            # WS set in TCPConnect having IP as third argument
            if not ("workstation_name" in signOnType and len(signOnType["workstation_name"]["byValueCount"]) == 1 and singleValue(signOnType, "workstation_name") == "10"):
                return False
                
            # LOA == 1 => CAPRI
            if not ("loa" in signOnType and len(signOnType["loa"]["byValueCount"]) == 1 and singleValue(signOnType, "workstation_name") == "1"):
                return False
                
            # Do warning on IP Addresses
                
            return True

        if "creator" in userInfo:
            self.__result["remoteExcludeReasonCount"]["NO_CREATOR"] += 1
            return False
            
        ALLOWED_PROPS = UserClassifier.PUREZEROSOCREATEDUSER_200_MAND_PROPS.union(UserClassifier.PUREZEROSOCREATEDUSER_200_OPT_PROPS)
        userInfo["hasProps"] = list(set(userInfo["hasProps"]) - set(["_id", "label", "type"]))
        if len(set(userInfo["hasProps"]) - ALLOWED_PROPS) > 0:
            self.__result["remoteExcludeReasonCount"]["ALLOWED_PROPS"] += 1
            return False
            
        # May be too strict: SMO + TimeStamp missing from some (subgroup?)
        # ... all seem to have LOA 1
        if len(UserClassifier.PUREZEROSOCREATEDUSER_200_MAND_PROPS - set(userInfo["hasProps"])):
            self.__result["remoteExcludeReasonCount"]["MAND_PROPS"] += 1
            missingMands = (UserClassifier.PUREZEROSOCREATEDUSER_200_MAND_PROPS - set(userInfo["hasProps"])) - set(["secondary_menu_options", "timestamp", "alias"])
            if len(missingMands) != 0:
                self.__result["warningsByUserRef"][userRef].append("At this point in Remote Check, only expected mandatory SMO/ TimeStamp/ Alias could be missing but {}".format("/".join(list(missingMands))))
            return False
            
        if "ssn" not in userInfo and sum(1 for vf in userInfo["signed_on_froms"] if re.match(r'358\:', vf)) == 0:
            self.__result["remoteExcludeReasonCount"]["MANILLA_NO_SSN"] += 1
            return False
            
        if "signed_on_froms" not in userInfo or sum(1 for sof in userInfo["signed_on_froms"] if re.match(self.__stationNo, sof)) == len(userInfo["signed_on_froms"]):
            self.__result["remoteExcludeReasonCount"]["ONLY_LOCAL_STATION_NO"] += 1
            return False
            
        if "date_entered" not in userInfo:
            warning = "No User 'date entered'"
            self.__result["warningsByUserRef"][userRef].append(warning)            
            
        # device or not: seems some have, some not - cause dropped later or was
        # earlier
                    
        mds = manyDivisionSummary(signOnType)
        if mds: # not printing as too many remote guys
            warning = "Not Default Division Only: {}".format(mds)
            self.__result["warningsByUserRef"][userRef].append(warning)
            
        return True
        
    """
    Parallel a strict remote with a pure Local and put others in Mixed as before
    
    TODO: MUCH too broad as have mixed remotes (just fail on not 0 created etc)
    
     * device: add mix of 0 and SSH (may split?) [enforce NOT for others?]
     * division selection ie/ ALake vs Default 
     * specific user activity such as creating appointments (clerks)
    """
    def __testLocalUser(self, userRef, userInfo, signOnType):
    
        if "is_created_by_0" in userInfo:
            warning = "0 CREATED"
            self.__result["warningsByUserRef"][userRef].append(warning) 
                       
        return True # take all for now
   
"""
Heuristic - use for warnings etc
"""     
def couldBeMachineSSN(ssn):
    if ssn in ["123456789"]:
        return True
    # >= 4 identical digits in a row
    # 000006777, 111116777, 099999999, 000998888, 123456789, 000005678 ...
    if re.search(r'(\d)\1{3,}', ssn):
        return True
    # Pairs RE
    if len(re.findall(r'(\d)\1{1,}', ssn)) >= 2:
        return True
    return False
    
def enforceLOA(signOnType, loaValue):
    loa = None
    try:
        loa = singleValue(signOnType, "level_of_assurance")
    except:
        return False
    if loa != loaValue:
        return False
    return True
    
def manyDivisionSummary(signOnType):
    if "division" not in signOnType or len(signOnType["division"]["byValueCount"]) == 1:
        return ""
    sdivMU = ", ".join(["{}:{}".format(div[0:6], signOnType["division"]["byValueCount"][div]) for div in signOnType["division"]["byValueCount"]])
    return sdivMU

    



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict
from datetime import datetime
import numpy

from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL, metaOfVistA
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent

from ..webReportUtils import ensureWebReportLocations, SITE_DIR_TEMPL, TOP_MD_TEMPL, keyStats

"""
Basic overview of Data available in system (suits a PDF/ all is for Web)
"""
def webReportDataTypesSummary(stationNo, typeInfoById=None):
       
    if typeInfoById == None:
        typeInfoById = expectedTypeCounts(stationNo)
    noRecords = sum(typeInfoById[typId]["count"] for typId in typeInfoById)

    title = "{} Data Types Summary".format(stationNo)
    mu = TOP_MD_TEMPL.format("Data Types Summary", title)
        
    mu += muIntroduction(stationNo, typeInfoById)
    
    mu += "The top 15 types are a mixture of patient record and system log and configuration data ...\n\n"
    
    top15 = set()
    tbl = MarkdownTable(["Rank", ":Type", "Records", "Share"], includeNo=False)
    for i, typId in enumerate(sorted(typeInfoById, key=lambda x: typeInfoById[x]["count"], reverse=True), 1):
        top15.add(typId)
        fid = re.sub(r'\_', '.', typId)
        cntMU = typeInfoById[typId]["count"] if typeInfoById[typId]["count"] != -1 else "UNTRACKED"
        percMU = reportPercent(typeInfoById[typId]["count"], noRecords) if typeInfoById[typId]["count"] > 0 else ""
        tbl.addRow([typeInfoById[typId]["rank"], "__{}__ ({})".format(typeInfoById[typId]["label"], fid), cntMU, percMU])
        if i == 15:
            break
    mu += tbl.md() + "\n\n"
    
    # In case not in the top 15
    OTHER_SIGNIFICANT_FILES = set([
        "3_081",
        "2", # PATIENT
        "9_4", # PACKAGE
        "44", # HOSPITAL LOCATION
        "200", # NEW PERSON
        "409_84", # SDEC APPOINTMENT
        "627_8", # DIAGNOSTIC RESULTS - MENTAL HEALTH 
        "631", # H... Patient
        "665", # Prosthetic Patient
        "790", # WV Patient
        "2005", # IMAGE
        "8925", # TIU DOCUMENT
        "8925_1", # TIU DOCUMENT DEFINITION
        "8994", # REMOTE PROCEDURE
    ])
        
    mu += "Other less populated but significant files include ...\n\n"
    
    tbl = MarkdownTable(["Rank", ":Type", "Records", "Share"], includeNo=False)
    for i, typId in enumerate(sorted(typeInfoById, key=lambda x: typeInfoById[x]["count"], reverse=True), 1):
        if typId not in OTHER_SIGNIFICANT_FILES:
            continue
        if typId in top15:
            continue
        fid = re.sub(r'\_', '.', typId)
        cntMU = typeInfoById[typId]["count"] if typeInfoById[typId]["count"] != -1 else "UNTRACKED"
        percMU = reportPercent(typeInfoById[typId]["count"], noRecords) if typeInfoById[typId]["count"] > 0 else ""
        tbl.addRow([typeInfoById[typId]["rank"], "__{}__ ({})".format(typeInfoById[typId]["label"], fid), cntMU, percMU])
    mu += tbl.md() + "\n\n"
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "dataTypesSummary.md", "w").write(mu)

"""
Full enumeration of available data (web suitable - summary suits a PDF)
"""    
def webReportDataTypesAll(stationNo, typeInfoById=None):

    if typeInfoById == None:
        typeInfoById = expectedTypeCounts(stationNo)
    noRecords = sum(typeInfoById[typId]["count"] for typId in typeInfoById)

    title = "{} Data Types".format(stationNo)
    mu = TOP_MD_TEMPL.format("Data Types", title)
    
    mu += muIntroduction(stationNo, typeInfoById)
    
    currentPercentile = -1
    tbl = None
    for i, typId in enumerate(sorted(typeInfoById, key=lambda x: typeInfoById[x]["count"], reverse=True), 1):
        fid = re.sub(r'\_', '.', typId)
        cntMU = typeInfoById[typId]["count"] if typeInfoById[typId]["count"] != -1 else "UNTRACKED"
        percMU = reportPercent(typeInfoById[typId]["count"], noRecords) if typeInfoById[typId]["count"] > 0 else ""
        sizePercentile = typeInfoById[typId]["sizePercentile"] if "sizePercentile" in typeInfoById[typId] else ""
        if sizePercentile != currentPercentile:
            currentPercentile = sizePercentile
            if tbl:
                mu += tbl.md() + "\n\n"
            mu += "{} Percentile ...\n\n".format(currentPercentile)
            tbl = MarkdownTable(["\#", ":Type", "Records", "Share"], includeNo=False)
            j = 1
        else:
            j += 1
        if typeInfoById[typId]["count"] == -1:
            if currentPercentile != -1:
                currentPercentile = -1
                tbl.addSeparatorRow()
            tbl.addRow(["{}/{}".format(j, i), "__{}__ ({})".format(typeInfoById[typId]["label"], fid), "", "", ""])    
            continue 
        tbl.addRow([i, "__{}__ ({})".format(typeInfoById[typId]["label"], fid), cntMU, percMU])
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "dataTypes.md", "w").write(mu)
    
"""
Basic summary shared by ALL and Summary reports
"""
def muIntroduction(stationNo, typeInfoById):

    noEmptyTypes = sum(1 for typId in typeInfoById if typeInfoById[typId]["count"] == 0)
    # untrack is -1
    noUntrackedTypes = sum(1 for typId in typeInfoById if typeInfoById[typId]["count"] == -1)
    noRecords = sum(typeInfoById[typId]["count"] for typId in typeInfoById)
    
    cnts = [typeInfoById[typId]["count"] for typId in typeInfoById if typeInfoById[typId]["count"] > 1]
    singletonCnts = sum(1 for typId in typeInfoById if typeInfoById[typId]["count"] == 1)
    kstats = keyStats(cnts)
    median = int(round(kstats["median"]))
    recordCntBiggest = sorted(cnts, reverse=True)[0]
    
    # 'fails' gracefully on meta and perhaps run BEFORE any but schema exist
    meta = metaOfVistA(stationNo)
    mu = "__{} [{}]__ was cut on _{}_ and reports <span class='yellowIt'>{:,}</span> types of data (\"File Types\"), <span class='yellowIt'>{}</span> of which are singletons (have only one record), <span class='yellowIt'>{}</span> are empty and <span class='yellowIt'>{:,}</span> are untracked. The system has a total of <span class='yellowIt'>{:,}</span> records. While the biggest type has <span class='yellowIt'>{:,}</span> records, the median number of records for types with more than one record is a lowly <span class='yellowIt'>{:,}</span>.\n\n".format(
        meta.get("name", "VistA"), 
        stationNo, 
        meta["cutDate"],
        len(typeInfoById), 
        reportAbsAndPercent(singletonCnts, len(typeInfoById)), 
        reportAbsAndPercent(noEmptyTypes, len(typeInfoById)), 
        noUntrackedTypes, 
        noRecords, 
        recordCntBiggest, 
        median
    )
    
    return mu
    
"""
Utility that using SELECT TYPES to rank files by count and give them
a size percentile

TODO: if full and SO for types then may add date summary ie/ first entry etc
"""
def expectedTypeCounts(stationNo):
    try:
        selectTypes = json.load(open(SCHEMA_LOCN_TEMPL.format(stationNo) + "SELECT_TYPES.json"))
    except:
        raise Exception("Can't load SELECT_TYPE.json from {}".format(SCHEMA_LOCN_TEMPL.format(stationNo)))
    expectedCountByType = {}
    ranks = set()
    cnts = []
    for result in selectTypes["results"]:
        if "parent" in result:
            continue
        typ = re.sub(r'\.', '_', result["number"])
        if "count" not in result:
            cnt = 0
        elif re.search(r'\-', result["count"]):
            cnt = -1
        else:
            cnt = int(result["count"]) 
            cnts.append(cnt)
        if cnt not in ranks:
            ranks.add(cnt)
        expectedCountByType[typ] = {"label": result["name"], "count": cnt}
    ranks = sorted(list(ranks), reverse=True)
    percentileThress = []
    for ptile in range(90, 0, -10):
        thres = numpy.percentile(cnts, ptile)
        percentileThress.append((int(thres), ptile))
    for typ in expectedCountByType:
        if expectedCountByType[typ]["count"] == -1: # no rank given
            continue
        expectedCountByType[typ]["rank"] = ranks.index(expectedCountByType[typ]["count"]) + 1
        if expectedCountByType[typ]["count"] == 0:
            continue
        for percThres in percentileThress:
            if expectedCountByType[typ]["count"] >= percThres[0]:
                expectedCountByType[typ]["sizePercentile"] = percThres[1]
                break
    return expectedCountByType

# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 4)
    
    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")
        
    ensureWebReportLocations(stationNo)

    webReportDataTypesSummary(stationNo)
                 
    webReportDataTypesAll(stationNo)
    
if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime
import platform
import statistics

from fmqlutils import VISTA_DATA_BASE_DIR
from fmqlutils.cacher.cacherUtils import SCHEMA_LOCN_TEMPL 
from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlreports.webReportUtils import vistasOfVISNByOne

"""
# TODO: move fmqr/directory

TODO WORK OFF/IMPROVE CACHER:
- create date:
  - make sure create date seen (map out # with a create date ie/ adding in fmqlreports meta) to make it clear
  - not updated since ie/ last X entries 
- fmql data qa ie/ to tighten that by exposing it here

TODO WORK (and delete) Archive:
- https://github.com/caregraf/cgArchive/blob/master/Voldermort/schemaRankReporter.py ... schema rank, word soup etc
- Voldermort/vistaAnalytics.py ... and buildInfos
"""

class CommonFileDirectory(): 

    ENTRY_IDX_SNO = 0

    # SELECT TYPES
    ENTRY_IDX_ID = 1
    ENTRY_IDX_NAME = 2
    ENTRY_IDX_PARENTID = 3
    ENTRY_IDX_COUNT = 4
    
    # DESCRIBE TYPE
    # TODO: expand with 
    # - count pointers, literals etc
    # - boolean on pointers to PATIENT etc
    ENTRY_IDX_FIELDCOUNT = 5

    def __init__(self, vistaSNOs):
        self.__vistaSNOs = vistaSNOs
        self.__fdirFile = f'{VISTA_DATA_BASE_DIR}Common/Directory/FDIR.json'
        try:
            print("Attempting to load preexisting file Directory")
            self.__fdirInfo = json.load(open(self.__fdirFile))
        except:
            print("Starting Directory from Scratch")
            self.__fdirInfo = {"vistas": [], "entries": {}, "files": [], "sentries": {}, "sfiles": []}
        else:
            print(f'Loaded preexisting directory with {len(self.__fdirInfo["files"]):,} files')
        self.__changeUpdate()
        
    def vistas(self):
        return self.__fdirInfo["vistas"]
                
    # Fan out the file - good for reports
    def lookupFileEntriesByGIEN(self, gien):
        gienFile = None
        for file in self.__fdirInfo["files"]:
            if gien in file:
                gienFile = file
                break
        if not gienFile:
            return None
        entries = []
        for giene in gienFile:
            entry = self.__fdirInfo["entries"][giene]
            entries.append(entry)
        return {"gien": gien, "entries": entries}
                
    def lookupEntryByGIEN(self, gien):
        return self.__fdirInfo["entries"].get(gien, None)
        
    """
    Report on contents
    ... basis for .md's and graphics both PER SYSTEM and CROSS
    
    Issue of >0, -1 (UNTRACKED), ...
    
    Do 
        “{var1} {var2}”.format(var1=value1, var2=value2) ie/ easier read
        
    TODO ADDITIONS:
    - DSS and other vendors break out and not just exclusive files and per station files
    - excluding the empties post the first pass as deprecated
    """
    def report(self):        
        
        mu = "# File Directory Synopsis\n\n"
        mu += f'VistAs: {len(self.__fdirInfo["vistas"]):,}' + "\n"
        
        # Graphics - top: topTotal - isHasBH
        fls = self.fileIds()
        mu += f'Files: {len(fls):,}' + "\n"
        cores = set(self.commonFiles())
        mu += "Core (in all): {}\n".format(
            reportAbsAndPercent(len(cores), len(fls))
        )
        commonCount = self.commonCounts()
        mu += f'Common Files: {"/".join([reportAbsAndPercent(commonCount[cnt], len(fls)) for cnt in sorted(commonCount) if cnt > 1])}' + "\n"
        emptiesInAll = set(self.filesOfCount(count=0))
        flsNotEmptyInAll = set(fls) - emptiesInAll
        commonCountNoEmpties = self.commonCounts(excludeFLs=emptiesInAll)
        mu += f'Common Files (w/o empties in all): {"/".join([reportAbsAndPercent(commonCountNoEmpties[cnt], len(flsNotEmptyInAll)) for cnt in sorted(commonCountNoEmpties)])}' + "\n"
        exclFiles = self.commonFiles(count=1) # if they appear, are one
        mu += "Exclusive (in 1): {}\n".format(
            reportAbsAndPercent(len(exclFiles), len(fls)),
        )
        mu += "Exclusive files - own station more: {}\n".format(
            reportAbsAndPercent(len(self.stationSpecificFiles()), len(exclFiles))
        )
        mu += "Files with > 1 name: {}\n".format(
            reportAbsAndPercent(len(self.filesWithManyNames()), len(fls))
        )

        totalRecords = self.totalRecords()
        mu += f'Records: {totalRecords:,}' + "\n"         
        allFlsByGroup = {}
        for fl in exclFiles:
            allFlsByGroup[fl] = "EXCLUSIVE"
        for fl in cores:
            allFlsByGroup[fl] = "ALL"
        for fl in fls:
            if fl in allFlsByGroup or fl in emptiesInAll:
                continue
            allFlsByGroup[fl] = "SOME"
        totalsByGroup = self.totalRecords(groupOfFile=allFlsByGroup)
        mu += f'Records (by group): {"/".join([reportAbsAndPercent(totalsByGroup[grp], totalRecords) for grp in sorted(totalsByGroup, key=lambda x: totalsByGroup[x], reverse=True)])}' + "\n"
        singletons = set(self.filesOfCount(count=1))
        mu += "Singletons: {}\n".format(
            reportAbsAndPercent(len(singletons), len(fls))
        )
        mu += "Singleton Cores: {}\n".format(
            reportAbsAndPercent(len(singletons.intersection(cores)), len(singletons))
        )
        # May later (add note to see - empty in any?)
        mu += "Empties: {}\n".format(
            reportAbsAndPercent(len(emptiesInAll), len(fls))
        )
        # May change focus to "populated" cores (or will triple with > 1 populated)
        mu += "Empty Cores: {}\n".format(
            reportAbsAndPercent(len(emptiesInAll.intersection(cores)), len(emptiesInAll))
        )
        mu += "Empty Exclusives: {}\n".format( # rem: cores same for all!
            reportAbsAndPercent(len(emptiesInAll.intersection(exclFiles)), len(emptiesInAll))
        )
        mu += "Untracked: {}\n".format(
            reportAbsAndPercent(len(self.untrackedFiles()), len(fls))
        )       
        mu += f'Definitions across VistAs: {len(self.__fdirInfo["entries"]):,}' + "\n"
        mu += "\n"
        for vistaInfo in self.__fdirInfo["vistas"]:
        
            stationNo = vistaInfo["stationNumber"]
            mu += f'## VistA {stationNo}\n' + "\n"
            
            fls = self.fileIds(stationNo)
            mu += f'Files: {len(fls):,}' + "\n"
            commonCount = self.commonCounts(stationNo)
            mu += f'Common Files: {"/".join([reportAbsAndPercent(commonCount[cnt], len(fls)) for cnt in sorted(commonCount) if cnt > 1])}' + "\n"
            commonCountNoEmpties = self.commonCounts(stationNo, emptiesInAll)
            mu += f'Common Files (w/o empties in all): {"/".join([reportAbsAndPercent(commonCountNoEmpties[cnt], len(flsNotEmptyInAll)) for cnt in sorted(commonCountNoEmpties)])}' + "\n"
            exclFiles = set(self.commonFiles(stationNo, count=1))
            mu += "Exclusive (in 1): {}\n".format(
                reportAbsAndPercent(len(exclFiles), len(fls)),
            )
            vistaSpecificFiles = self.stationSpecificFiles(stationNo, [stationNo])
            mu += "Exclusive files - own station more: {}\n".format(
                reportAbsAndPercent(
                    len(vistaSpecificFiles),
                    len(exclFiles)
                )
            )
            mu += f'Records: {self.totalRecords(stationNo):,}' + "\n"
            totalsByGroup = self.totalRecords(stationNo, groupOfFile=allFlsByGroup)
            mu += f'Records (by group): {"/".join([reportAbsAndPercent(totalsByGroup[grp], totalRecords) for grp in sorted(totalsByGroup, key=lambda x: totalsByGroup[x], reverse=True)])}' + "\n"
            mu += f'Median Record per file (inc empties): {self.medianRecords(stationNo, True):,}' + "\n"
            mu += f'Median Record per file (excl empties): {self.medianRecords(stationNo, False):,}' + "\n"
            singletons = set(self.filesOfCount(stationNo, count=1))
            mu += "Singletons: {}\n".format(
                reportAbsAndPercent(
                    len(singletons),
                    len(fls)
                )
            )
            mu += "Singleton Cores: {}\n".format( # rem: cores same for all!
                reportAbsAndPercent(len(singletons.intersection(cores)), len(singletons))
            )
            empties = set(self.filesOfCount(stationNo, count=0))
            mu += "Empties: {}\n".format(
                reportAbsAndPercent(
                    len(empties),
                    len(fls)
                )
            )
            mu += "Empty Cores: {}\n".format( # rem: cores same for all!
                reportAbsAndPercent(len(empties.intersection(cores)), len(empties))
            )
            mu += "Empty Exclusives: {}\n".format( # rem: cores same for all!
                reportAbsAndPercent(len(empties.intersection(exclFiles)), len(empties))
            )
            mu += "Untracked: {}\n".format(
                reportAbsAndPercent(
                    len(self.untrackedFiles(stationNo)),
                    len(fls)
                )
            )
            mu += "\n" 
                
        return mu
        
    # ----------------- File Ids and Presence across VistAs -------------------
            
    def fileIds(self, stationNo=""):
        """
        All files known or just per VistA
        
        For Plot: enables a isHasBH - total files and then can give % with different 
        attributes using other methods below (emptyFileIds)
        """
        fids = [fl[0].split(":")[1] for fl in self.__fdirInfo["files"]] if stationNo == "" else [entry[CommonFileDirectory.ENTRY_IDX_ID] for entry in self.__entriesByVistA(stationNo)]
        return sorted(fids, key=lambda x: float(x))
        
    def commonFiles(self, stationNo="", count=0): 
        """
        # 0 == ALL
            ie/ commonFiles(stationNo) <=> allFiles([stationNo])
        # 1 == Exclusive (ie/ not common at all)
            ie/ commonFiles(stationNo, 1) <=> exclusiveFiles([stationNo])
        where stationNo being set limits the list to files in a particular vista
        
        # > 1 BUT < len(vistas) => mix ie/ not exclusive and not all
        """
        inCount = []
        count = count or len(self.__fdirInfo["vistas"])
        for fl in self.__fdirInfo["files"]:
            if len(fl) != count:
                continue
            if stationNo and sum(1 for f in fl if f.split(":")[0] != stationNo):
                continue
            inCount.append(fl[0].split(":")[1])
        return inCount
        
    def filesWithManyNames(self):
        manyNamedFls = []
        for fl in self.__fdirInfo["files"]:
            nms = set(self.__fdirInfo["entries"][f][CommonFileDirectory.ENTRY_IDX_NAME] for f in fl)
            if len(nms) > 1:
                manyNamedFls.append(fl[0].split(":")[1])
        return manyNamedFls
        
    def stationSpecificFiles(self, stationNo="", stationsToMatch=None):
        """
        - stationsToMatch: if None => vistas know in directory
        
        Picks out files with ids in the station space ie/ if 668 is station
        then file id starts with 668
        
        ex 687000, 6870500 ... 663 goes further 6631965
        """
        inSpecifics = []
        if stationsToMatch == None:
            stationsToMatch = [vi["stationNumber"] for vi in self.__fdirInfo["vistas"]]
        stationsToMatchRE = "({})".format("|".join(stationsToMatch))
        for fl in self.__fdirInfo["files"]:        
            if stationNo and sum(1 for f in fl if f.split(":")[0] != stationNo):
                continue
            if not re.match(stationsToMatchRE, fl[0].split(":")[1]):
                continue
            fid = fl[0].split(":")[1]
            if len(fid) < 6:
                continue # 668000 on
            inSpecifics.append(fid)
        return inSpecifics
                
    def commonCounts(self, stationNo="", excludeFLs=set()):
        """
        By count of vista's types are in, count types ex/
            4: 500 <---- in 4 VistAs
            3: 400 <---- in 3 VistAs
            ...
        
        Arguments:
        - stationNo: only do for files of one VistA
        - excludeFLs: can exclude files (ex/ empty files or ...) so 
        count files of different sorts

        For Plot: the SchemaCommon/plotCategoryBSV Plot and Explanation
        """
        crossVistATypesPresenceCount = Counter()
        for fl in self.__fdirInfo["files"]:
            crossVistATypesPresenceCount[fl[0].split(":")[1]] = len(fl)
        fls = self.fileIds(stationNo) # all if stationNo == ""
        commonCount = Counter()
        for flId in fls:
            if flId in excludeFLs:
                continue
            commonCount[crossVistATypesPresenceCount[flId]] += 1                    
        return commonCount
        
    # ----------------- Record Count and Categorize by Record Count -------------------
           
    def totalRecords(self, stationNo="", groupOfFile=None):
        """
        Support any grouping of files: fl: GRP ... ex/ Patient vs ...
        or Exclusive files vs ...
        
        For Plot: categoryBH (ala AllConsults top of IFC)
    
            One Bar broken down ...
    
            plotData["allRecords"] = {
                "title": "All Records",
                "plotName": "fileAllRecords",
                "plotMethod": "plotCategoryBH",
                "rows": ["records"],
                "columns": [],
                "data": []
            }
            dataOne = []
            for vistaInfo in self.__fdirInfo["vistas"]:
                stationNo = vistaInfo["stationNumber"]
                plotData["columns"].append(stationNo)
                dataOne.append(self.totalRecords(stationNo))
            plotData["data"].append(dataOne)
        """
        ttl = 0
        ttlByGRP = Counter()
        for gid in self.__fdirInfo["entries"]:
            if stationNo and not re.match(stationNo, gid):
                continue
            cnt = self.__fdirInfo["entries"][gid][CommonFileDirectory.ENTRY_IDX_COUNT]
            if cnt in [0, -1]:
                continue  
            if groupOfFile:
                flId = gid.split(":")[1]
                if flId not in groupOfFile:
                    raise Exception(f"Expect all files to be id'ed in groupOfFile: {flId}")
                grp = groupOfFile[flId]
                ttlByGRP[grp] += cnt                 
                continue 
            ttl += cnt
        return ttl or ttlByGRP

    def filesOfCount(self, stationNo="", count=-1):
        """
        - Empty files: count = 0
        - Singleton files: count = 1
        
        TODO: consider bucketing ie/ count=>bounds ie/ > 10 etc
        """
        flCounts = defaultdict(list)
        for entry in self.__fdirInfo["entries"].values():
            if stationNo and entry[CommonFileDirectory.ENTRY_IDX_SNO] != stationNo:
                continue
            fl = entry[CommonFileDirectory.ENTRY_IDX_ID]
            cnt = entry[CommonFileDirectory.ENTRY_IDX_COUNT]
            flCounts[fl].append(cnt)
        fls = []
        for fl in flCounts:
            if len(set(flCounts[fl])) > 1 or flCounts[fl][0] != count:
                continue   
            fls.append(fl) 
        if "996.2" in fls:
            print(flCounts["996.2"])
        return fls
        
    def untrackedFiles(self, stationNo=""):
        """
        Untracked files
        """
        return self.filesOfCount(stationNo, count=-1)
        
    def medianRecords(self, stationNo="", includeEmpties=True):
        """
        Arguments:
        - StationNo: "" => all known systems ie/ if merged; else one VistA's files
        - includeEmpties: ignore empty files (or not)
        ... always excludes untracked files
        """
        entries = self.__fdirInfo["entries"].values() if stationNo == "" else self.__entriesByVistA(stationNo)        
        minToCount = 0 if includeEmpties else 1
        medianRecordsPerFile = statistics.median([
            entry[CommonFileDirectory.ENTRY_IDX_COUNT]
            for entry in entries if entry[CommonFileDirectory.ENTRY_IDX_COUNT] >= minToCount
        ]) # don't count the untracked
        return medianRecordsPerFile

    # -------------------------- Utilities etc --------------------------
        
    def __entriesByVistA(self, stationNo):
        return [self.__fdirInfo["entries"][gid] for gid in self.__fdirInfo["entries"] if re.match(stationNo, gid)]
        
    # TODO: timing stuff is ever get to 9_4, 9_6 - see old Archives
             
    # ########################## CHANGE / UPDATE ######################
                        
    def __changeUpdate(self):
        changed = self.__fillFromAvailable()
        if changed:
            print("Changed - so rebuilding files")
            self.__rebuildFiles()
            self.__flush()    
                             
    # Only Select Types for now - can expand later   
    def __fillFromAvailable(self):
    
        """
        string.isnumeric() Returns True if there are only numeric characters in the string. If not, returns False.
string.isalpha() Returns True if there are only alphabetic characters in the string. If not, returns False.
        """
        def processSelectTypesOfStation(stationNo):
            try:
                selectTypes = json.load(open(SCHEMA_LOCN_TEMPL.format(stationNo) + "SELECT_TYPES.json"))
            except:
                raise Exception("Can't load SELECT_TYPE.json from {}".format(SCHEMA_LOCN_TEMPL.format(stationNo)))
            if selectTypes["siteId"] != stationNo:
                raise Exception(f'Didn\'t get SELECT TYPES expected for {stationNo}')
            print(f'Extracting entries from SELECTTYPES of {selectTypes["siteLabel"]}') 
            entries = []
            sentries = []
            for result in selectTypes["results"]:
                entry = [
                    stationNo, # STATION NO
                    result["number"], # ID
                    result["name"], # NAME
                ]
                if "parent" in result:
                    entry.extend([result["parent"], -1])
                    sentries.append(entry)
                    continue
                if "count" not in result:
                    cnt = 0
                elif re.search(r'\-', result["count"]):
                    cnt = -1
                else:
                    cnt = int(result["count"]) 
                entry.extend(["", cnt])  
                entries.append(entry)
            print(f'\tbuilt {len(entries):,} entries')
            return entries, sentries     
            
        changed = False
        for stationNo in self.__vistaSNOs:
            if sum(1 for vistaInfo in self.__fdirInfo["vistas"] if vistaInfo["stationNumber"] == stationNo):
                print("\t{} already in Directory - skipping".format(stationNo))
                continue
            try:
                stationEntries, stationSEntries = processSelectTypesOfStation(stationNo)
            except:
                print(f'Can\'t load SelectTypes for {stationNo} as not available/creatable here')
                raise
            changed = True
            for stationEntry in stationEntries:
                self.__fdirInfo["entries"][
                f'{stationNo}:{stationEntry[CommonFileDirectory.ENTRY_IDX_ID]}'] = stationEntry
            for stationSEntry in stationSEntries:
                self.__fdirInfo["sentries"][
                f'{stationNo}:{stationSEntry[CommonFileDirectory.ENTRY_IDX_ID]}'] = stationSEntry
            self.__fdirInfo["vistas"].append({"stationNumber": stationNo, "dateAdded": str(datetime.now()).split(" ")[0]})
        return changed
                
    def __rebuildFiles(self): # Just doing top files for now
        del self.__fdirInfo["files"]
        fileEntryGIDsById = defaultdict(list)
        for gid in self.__fdirInfo["entries"]:
            entry = self.__fdirInfo["entries"][gid]
            fileEntryGIDsById[entry[CommonFileDirectory.ENTRY_IDX_ID]].append(
                f'{entry[CommonFileDirectory.ENTRY_IDX_SNO]}:{entry[CommonFileDirectory.ENTRY_IDX_ID]}'
            )
        self.__fdirInfo["files"] = [fileEntryGIDsById[fid] for fid in sorted(fileEntryGIDsById, key=lambda x: float(x))]
        print(f'From {len(self.__fdirInfo["entries"]):,}, isolated {len(self.__fdirInfo["files"]):,} distinct files')

    def __flush(self):
        print("Files {:,}, Entries {:,} flushed".format(len(self.__fdirInfo["files"]), len(self.__fdirInfo["entries"])))
        json.dump(
            self.__fdirInfo, 
            open(self.__fdirFile, "w"), 
            indent=4
        )
        
# ############################## DRIVER ##########################
                
def main():

    assert sys.version_info >= (3, 6)
    
    cfd = CommonFileDirectory(vistasOfVISNByOne("20"))
    print(cfd.report())
            
if __name__ == "__main__":
    main() 


#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from collections import defaultdict, Counter
from datetime import datetime

from fmqlutils.reporter.reportUtils import MarkdownTable, reportPercent, reportAbsAndPercent
from fmqlutils.typer.reduceTypeUtils import splitTypeDatas, checkDataPresent, singleValue
from ..webReportUtils import filteredResultIterator, TOP_MD_TEMPL, SITE_DIR_TEMPL, keyStats, roundFloat

"""
Health Factors Basic

Examine categories types more (http://www.cbo.gov/sites/default/files/cbofiles/ftpdocs/88xx/doc8892/maintext.3.1.shtml, 

Requires:
- 9999999_64 ALL (Filtered Walk)
- 9000010_23 (SOReduction/health_factor)

TODO v1.1: 
- 9000010_23 SPO has few but still present invalid visit dates of YR-MTH-00 so excluded (filtered) by typer
- put in last signon for everyone mentioned (do same for Dental)

And note that in PCE

    Patient Care Encounter (PCE) helps sites collect, manage, and display outpatient
    encounter data (including providers, procedure codes, and diagnostic codes) in
    compliance with the 10/1/96 Ambulatory Care Data Capture mandate from the
    Undersecretary of Health.     (https://www.va.gov/vdl/documents/Clinical/Patient_Care_Encounter_(PCE)/pxtm.pdf)

TODO: see other xxrepo notes
"""
def webReportHealthFactors(stationNo):
        
    allThere, details = checkDataPresent(stationNo, [
        {"fileType": "9999999_64", "check": "ALL"},
        {"fileType": "9000010_23", "check": "TYPESO"} # subtype prop checked below
    ])
    if not allThere:
        raise Exception("Some required data is missing - {}".format(details))
        
    mu = """---
layout: default
title: {} Health Factors
--- 

""".format(stationNo) 

    # Definitions
    resourceIter = filteredResultIterator(stationNo, "9999999_64")
    redByRef = {}
    for resource in resourceIter: 
        # Not ideal - SPO has beyond ASCII - fix in Python 3
        label = resource["label"].encode('ascii', 'ignore').decode('ascii')
        resRef = "{} [{}]".format(re.sub(r'\_', '/', label), resource["_id"])
        red = {"ref": resRef}
        if "category" in resource:
            if resource["category"]["id"] == resource["_id"]:
                red["is_category_self"] = True
            else:
                categoryRef = "{} [{}]".format(re.sub(r'\_', '/', resource["category"]["label"]), resource["category"]["id"])
                red["category"] = categoryRef
        if "inactive_flag" in resource:
            red["is_inactive"] = True
        if "display_on_health_summary" in resource and resource["display_on_health_summary"]:
            red["is_display_on_health_summary"] = True
        if resource["entry_type"] == "F:FACTOR":
            red["is_factor"] = True
        redByRef[red["ref"]] = red
    factors = set(ref for ref in redByRef if "is_factor" in redByRef[ref])
    categories = set(ref for ref in redByRef if "is_factor" not in redByRef[ref])
    inactives = set(ref for ref in redByRef if "is_inactive" in redByRef[ref])
    actives = set(ref for ref in redByRef if "is_inactive" not in redByRef[ref])

    # Assignments
    all9000010_23Data, st9000010_23s = splitTypeDatas(stationNo, "9000010_23", reductionLabel="YR1", expectSubTypeProperty="health_factor")        
    usedFactors = set(re.sub(r'\_', '/', list(subTypeInfo["health_factor"]["byValueCount"])[0].split(" [")[0]) + " [" + list(subTypeInfo["health_factor"]["byValueCount"])[0].split(" [")[1] for subTypeInfo in st9000010_23s)
    subTypeInfoById = dict((re.sub(r'\_', '/', list(subTypeInfo["health_factor"]["byValueCount"])[0].split(" [")[0]) + " [" + list(subTypeInfo["health_factor"]["byValueCount"])[0].split(" [")[1], subTypeInfo) for subTypeInfo in st9000010_23s)
    usedFactorCounts = dict((factorRef, subTypeInfoById[factorRef]["_total"]) for factorRef in subTypeInfoById)
    usedCategoryCounts = Counter()
    usedFactorByCategoryCounts = defaultdict(lambda: Counter())
    factorsByCategory = defaultdict(list)
    for ref in redByRef:
        red = redByRef[ref]
        if "category" in red:
            factorsByCategory[red["category"]].append(ref)
    for factorRef in subTypeInfoById:
        red = redByRef[factorRef]
        if "category" in red:
            categoryRef = red["category"] 
        elif "is_factor" not in red:
            categoryRef = factorRef
        else:
            categoryRef = "** NO CATEGORY **"
        usedCategoryCounts[categoryRef] += subTypeInfoById[factorRef]["_total"]
        usedFactorByCategoryCounts[categoryRef][factorRef] = subTypeInfoById[factorRef]["_total"]

    overallFirstCreateDate = all9000010_23Data["_firstCreateDate"].split("T")[0]
    overallLastCreateDate = all9000010_23Data["_lastCreateDate"].split("T")[0]

    kstatsCategories = keyStats(usedCategoryCounts.values())
    maxCategoryLabel = sorted(usedCategoryCounts, key=lambda x: usedCategoryCounts[x], reverse=True)[0].split(" [")[0]
    kstatsFactors = keyStats([subTypeInfo["_total"] for subTypeInfo in st9000010_23s])
    maxFactorLabel = sorted(usedFactorCounts), key=lambda x: usedFactorCounts[x], reverse=True)[0].split(" [")[0]
    unassignedActiveFactors = factors - set(usedFactorCounts)
    unassignedActiveCategories = categories - set(usedCategoryCounts)
            
    mu += """<span class='yellowIt'>{:,}</span> health factor assignments were made in this VistA from {} through {}, the last full year for which data is available. The assignments cover <span class='yellowIt'>{:,}</span> health factors, grouped into <span class='yellowIt'>{:,}</span> categories for <span class='yellowIt'>{:,}</span> patients.
 Most nominally active factors - <span class='yellowIt'>{}</span> - and categories -  <span class='yellowIt'>{}</span> - are not assigned to a patient. They go _Unused_.
    
The median number of assignments per factor is <span class='yellowIt'>{:,}</span>, while the maximum is <span class='yellowIt'>{:,}</span> for factor _{}_. The median number per category is <span class='yellowIt'>{:,}</span>, while the maximum is <span class='yellowIt'>{:,}</span> for category _{}_.    
        
""".format(
        all9000010_23Data["_total"], 
        overallFirstCreateDate, 
        overallLastCreateDate,
        len(st9000010_23s), 
        len(usedCategoryCounts), 
        all9000010_23Data["patient_name"]["rangeCount"],
        
        reportAbsAndPercent(len(unassignedActiveFactors), len(factors.intersection(actives))), 
        reportAbsAndPercent(len(unassignedActiveCategories), len(categories.intersection(actives))), 
        
        roundFloat(kstatsFactors["median"]), 
        kstatsFactors["max"], 
        maxFactorLabel, 
        roundFloat( kstatsCategories["median"]), 
        kstatsCategories["max"], 
        maxCategoryLabel
    )
    
    categoriesAssigned = usedFactors.intersection(categories)
    if len(categoriesAssigned):
        mu += """__Note__: patients are typically assigned specific health factors, but in this system <span class='yellowIt'>{:,}</span> categories are assigned directly to patients.
        
""".format(len(categoriesAssigned))

    tbl = MarkdownTable(["Category", "Assignments", "Used Factors", "Unused Factors"])
    assignmentCount = 0
    for i, categoryRef in enumerate(sorted(usedCategoryCounts, key=lambda x: usedCategoryCounts[x], reverse=True), 1): 
        if i > 20:
            break
        unusedActiveFactors = set(factorsByCategory[categoryRef]).intersection(actives) - set(usedFactorByCategoryCounts[categoryRef])
        row = [
            "__{}__".format(categoryRef.split(" [")[0]), 
            reportAbsAndPercent(usedCategoryCounts[categoryRef], 
            all9000010_23Data["_total"]), 
            len(usedFactorByCategoryCounts[categoryRef]), 
            len(unusedActiveFactors)
        ]
        tbl.addRow(row)
        assignmentCount += usedCategoryCounts[categoryRef]
    mu += """The following shows the top 20 (out of <span class='yellowIt'>{:,}</span>) categories used for assigning health factors to patients. Collectively they account for <span class='yellowIt'>{}</span> of the system's assignments during this period. The table also shows how many of the factors of a category are used and how many go unused.
    
{}

""".format(
        len(usedCategoryCounts), 
        reportAbsAndPercent(assignmentCount, all9000010_23Data["_total"]), 
        tbl.md()
    )

    tbl = MarkdownTable(["Factor", "Category", "Assignments"])
    assignmentCount = 0
    for i, factorRef in enumerate(sorted(usedFactorCounts, key=lambda x: usedFactorCounts[x], reverse=True), 1): 
        if i > 20:
            break    
        red = redByRef[factorRef]
        factorLabel = factorRef.split(" [")[0]
        categoryMU = "__{}__".format(factorLabel) if "is_factor" not in red else red["category"].split(" [")[0]
        row = [
            "__{}__".format(factorLabel),
            categoryMU,
            reportAbsAndPercent(usedFactorCounts[factorRef], all9000010_23Data["_total"]) 
        ]
        tbl.addRow(row)
        assignmentCount += usedFactorCounts[factorRef]
    mu += """The following shows the top 20 (out of <span class='yellowIt'>{:,}</span>) health factors assigned to patients. Collectively they account for <span class='yellowIt'>{}</span> of the system's assignments. 
    
{}

""".format(
        len(usedFactorCounts), 
        reportAbsAndPercent(assignmentCount, all9000010_23Data["_total"]), 
        tbl.md()
    )
    
    mu += """__More to Analyze__:
    
  * __Clinical Reminders__ leverage Health Factors - hence the prevalence of _REMINDER FACTORS_. This needs to be fully examined.
  * Break out the Reminder Categories by broad topics to see just how much VA employs these factors for quite different activities - remind to educate, note homelessness, note an advance directive is refused.
  * The management of Health Factor and Category Definitions seems haphazard. Most go unused and the way they are marked inactive - make a factor inactive OR make its category inactive - seems inconsistent. This needs further examination.
  * The tie between documents (TIU Notes) and factors needs to be explored - in this system 99.8% of factors have TIU Notes as their data source.
  * Direct assignment of categories happens but it is unusual - is this because an individual factor type was turned into a category after assignment or because categories can be directly assigned.
  * Neither the definitions nor the assignments of this system have been standardized - standard codes go unused and factor definitions are not marked LOCAL or NATIONAL. A more recent system needs to be examined to see how far VA has gone with __Factor Standardization__. There are factors named VA- and these need to be examined further.
  

"""
    
    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    open(userSiteDir + "healthFactors.md", "w").write(mu)
    
# ################################# DRIVER #######################
               
def main():

    assert sys.version_info >= (3, 4)

    try:
        stationNo = sys.argv[1]
    except IndexError:
        raise SystemExit("Usage _EXE_ STATIONNO")

    userSiteDir = SITE_DIR_TEMPL.format(stationNo)
    if not os.path.isdir(userSiteDir):
        raise SystemExit("Expect User Site to already exist with its basic contents")

    webReportHealthFactors(stationNo)
                 
if __name__ == "__main__":
    main()
